import numpy as np
from pynwb.core import DynamicTable
from hdmf.common.table import VectorData
from nwbwidgets.utils.dynamictable import infer_categorical_columns



def test_infer_categorical_columns():
    
    data1 = np.array([1,2,2,3,1,1,3,2,3])
    data2 = np.array([3,4,2,4,3,2,2,4,4])
    
    vd1 = VectorData('Data1','vector data for creating a DynamicTable',data=data1)
    vd2 = VectorData('Data2','vector data for creating a DynamicTable',data=data2)
    vd=[vd1,vd2]
    
    dynamic_table = DynamicTable(name='test table',description='This is a test table',
                             columns=vd,colnames=['Data1','Data2'])
    
    assert all(infer_categorical_columns(dynamic_table))==all({'Data1': np.array([1, 2, 3]), 'Data2': np.array([2, 3, 4])})



import pickle


class MemoizeMutable:
    def __init__(self, fn):
        self.fn = fn
        self.memo = {}

    def __call__(self, *args, **kwds):
        this_str = pickle.dumps(args, 1) + pickle.dumps(kwds, 1)
        if this_str not in self.memo:
            self.memo[this_str] = self.fn(*args, **kwds)
        return self.memo[this_str]



from typing import Iterable

import numpy as np

from pynwb.misc import Units
import ipywidgets as widgets

from .misc import RasterWidget, PSTHWidget, RasterGridWidget
from .view import default_neurodata_vis_spec
from .utils.pynwb import robust_unique
from .controllers import GroupAndSortController


class AllenRasterWidget(RasterWidget):
    def make_group_and_sort(self, group_by=None):
        return AllenRasterGroupAndSortController(self.units, group_by=group_by)


class AllenPSTHWidget(PSTHWidget):

    def __init__(self, units: Units, unit_index=0, unit_controller=None, sigma_in_secs=.05, ntt=1000):

        super().__init__(units, unit_index, unit_controller, sigma_in_secs, ntt)

        self.stimulus_type_dd = widgets.Dropdown(options=np.unique(self.trials['stimulus_name'][:]).tolist(),
                                                 label='drifting_gratings',
                                                 description='stimulus type')

        self.stimulus_type_dd.observe(self.stimulus_type_dd_callback)

        self.children = [self.stimulus_type_dd] + list(self.children)

    def get_trials(self):
        return self.units.get_ancestor('NWBFile').epochs

    def stimulus_type_dd_callback(self, change):
        self.gas.discard_rows = np.where(self.trials['stimulus_name'][:] != self.stimulus_type_dd.value)[0]

    def make_group_and_sort(self, window=False):
        discard_rows = np.where(self.trials['stimulus_name'][:] != 'drifting_gratings')[0]
        gas = GroupAndSortController(self.trials, window=window, start_discard_rows=discard_rows)

        return gas


class AllenRasterGroupAndSortController(GroupAndSortController):

    def get_groups(self):

        self.electrodes = self.dynamic_table.get_ancestor('NWBFile').electrodes

        groups = super().get_groups()
        groups.update({name: np.unique(self.electrodes[name][:]) for name in self.electrodes.colnames})
        return groups

    def get_orderable_cols(self):
        units_orderable_cols = super().get_orderable_cols()
        candidate_cols = [x for x in self.electrodes.colnames
                          if not (isinstance(self.electrodes[x][0], Iterable) or
                                  isinstance(self.electrodes[x][0], str))]
        return units_orderable_cols + [x for x in candidate_cols
                                       if len(robust_unique(self.electrodes[x][:])) > 1]

    def get_group_vals(self, by, rows_select=()):
        if by is None:
            return None
        elif by in self.dynamic_table:
            return self.dynamic_table[by][:][rows_select]
        else:
            if self.electrodes is not None and by in self.electrodes:
                ids = self.electrodes.id[:]
                inds = [np.argmax(ids == val) for val in self.dynamic_table['peak_channel_id'][:]]
                return self.electrodes[by][:][inds][rows_select]


class AllenRasterGridWidget(RasterGridWidget):
    def get_trials(self):
        return self.units.get_ancestor('NWBFile').epochs

    def select_trials(self):
        self.controls['trials_select'] = widgets.Dropdown(options=np.unique(self.trials['stimulus_name'][:]).tolist(),
                                                          label='drifting_gratings',
                                                          description='trial select')
        self.children = list(self.children) + [self.controls['trials_select']]

    def process_controls(self, control_states):
        control_states['trials_select'] = self.trials['stimulus_name'][:] == control_states.pop('trials_select')
        return control_states


def load_allen_widgets():
    default_neurodata_vis_spec[Units]['Session Raster'] = AllenRasterWidget
    default_neurodata_vis_spec[Units]['Grouped PSTH'] = AllenPSTHWidget
    default_neurodata_vis_spec[Units]['Raster Grid'] = AllenRasterGridWidget




import matplotlib.pyplot as plt
import ipywidgets as widgets
import pynwb
from pynwb.image import GrayscaleImage, ImageSeries, RGBImage
from .base import fig2widget


def show_image_series(image_series: ImageSeries, neurodata_vis_spec: dict):

    def show_image(index=0):
        fig, ax = plt.subplots(subplot_kw={'xticks': [], 'yticks': []})
        ax.imshow(image_series.data[index, :, :], cmap='gray')
        plt.show()
        return fig2widget(fig)

    slider = widgets.IntSlider(value=0, min=0,
                               max=image_series.data.shape[0] - 1,
                               orientation='horizontal')
    controls = {'index': slider}
    out_fig = widgets.interactive_output(show_image, controls)
    vbox = widgets.VBox(children=[out_fig, slider])

    return vbox


def show_index_series(index_series, neurodata_vis_spec: dict):
    show_timeseries = neurodata_vis_spec[pynwb.TimeSeries]
    series_widget = show_timeseries(index_series)

    indexed_timeseries = index_series.indexed_timeseries
    image_series_widget = show_image_series(indexed_timeseries,
                                            neurodata_vis_spec)

    return widgets.VBox([series_widget, image_series_widget])


def show_grayscale_image(grayscale_image: GrayscaleImage, neurodata_vis_spec=None):
    fig, ax = plt.subplots()
    plt.imshow(grayscale_image.data[:], 'gray')
    plt.axis('off')

    return fig


def show_rbg_image(rgb_image: RGBImage, neurodata_vis_spec=None):
    fig, ax = plt.subplots()
    plt.imshow(rgb_image.data[:])
    plt.axis('off')

    return fig



import ipywidgets as widgets
from nwbwidgets.controllers import move_int_slider_down, move_int_slider_up, \
    move_range_slider_down, move_range_slider_up, move_slider_down, move_slider_up, RangeController
import unittest


def test_float_range_controller():
    assert isinstance(RangeController(vmin=1, vmax=26), widgets.Widget)

    
class MoveSliderTestCase(unittest.TestCase):

    def setUp(self):
        self.slider = widgets.IntSlider(
            value=7,
            min=2,
            max=10,
            step=1,
            description='Test:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d')

    def test_move_slider_down_bigger(self):
        move_slider_down(self.slider, 3)
        assert(self.slider.value == 4)

    def test_move_slider_down_smaller(self):
        move_slider_down(self.slider, 6)
        assert(self.slider.value == 2)

    def test_move_slider_up_bigger(self):
        move_slider_up(self.slider, 2)
        assert(self.slider.value == 8)

    def test_move_slider_up_smaller(self):
        move_slider_up(self.slider, 6)
        assert(self.slider.value == 4)

    def test_move_int_slider_down(self):
        move_int_slider_down(self.slider)
        assert(self.slider.value == 6)
        
    def test_move_int_slider_up(self):
        move_int_slider_up(self.slider)
        assert(self.slider.value == 8)


class RangeSliderTestCase(unittest.TestCase):

    def setUp(self):
        self.slider = widgets.IntRangeSlider(
            value=[5, 7],
            min=0,
            max=10,
            step=1,
            description='Test:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d')
        
    def test_move_range_slider_down_bigger(self):
        self.slider.value = (4, 6)
        move_range_slider_down(self.slider)
        assert(self.slider.value == (2, 4))

    def test_move_range_slider_down_smaller(self):
        self.slider.value = (2, 6)
        move_range_slider_down(self.slider)
        assert(self.slider.value == (0, 4))
    
    def test_move_range_slider_up_smaller(self):
        self.slider.value = (5, 7)
        move_range_slider_up(self.slider)
        assert(self.slider.value == (7, 9))

    def test_move_range_slider_up_bigger(self):
        self.slider.value = (5, 8)
        move_range_slider_up(self.slider)
        assert(self.slider.value == (7, 10))



from nwbwidgets import view
import matplotlib.pyplot as plt
from ipywidgets import widgets
from IPython import display
from collections.abc import Iterable
from pynwb import ProcessingModule
from pynwb.core import NWBDataInterface
from matplotlib.pyplot import Figure
from datetime import datetime
from typing import Union
import pandas as pd

GroupingWidget = Union[widgets.Accordion, widgets.Tab]


def show_fields(node, **kwargs) -> widgets.Widget:
    field_lay = widgets.Layout(max_height='40px', max_width='600px',
                               min_height='30px', min_width='130px')
    info = []
    for key, val in node.fields.items():
        lbl_key = widgets.Label(key+':', layout=field_lay)
        lbl_val = widgets.Label(str(val), layout=field_lay)
        info.append(widgets.HBox(children=[lbl_key, lbl_val]))
    vbox = widgets.VBox(info)
    return vbox


# def show_dynamic_table(node: DynamicTable, **kwargs):
def show_dynamic_table(node, **kwargs) -> widgets.Widget:
    out1 = widgets.Output()
    with out1:
        display.display(node.to_dataframe())
    return out1


def show_neurodata_base(node: NWBDataInterface, neurodata_vis_spec: dict) -> widgets.Widget:
    """
    Gets a pynwb object and returns a Vertical Box containing textual info and
    an expandable Accordion with it's children.
    """
    field_lay = widgets.Layout(max_height='40px', max_width='500px',
                               min_height='30px', min_width='180px')
    info = []         # string data type, exposed as a Text widget
    neuro_data = []   # more complex data types, also with children
    labels = []
    for key, value in node.fields.items():
        if isinstance(value, (str, datetime)):
            lbl_key = widgets.Label(key+':', layout=field_lay)
            lbl_val = widgets.Label(str(value), layout=field_lay)
            info.append(widgets.HBox(children=[lbl_key, lbl_val]))
        elif key == 'related_publications':
            pub_list = []
            for pub in value:
                pub_list.append(widgets.HTML(value="<a href=http://dx.doi.org/"+pub[4:]+">"+pub+"</a>"))
            lbl_key = widgets.Label(key+':', layout=field_lay)
            pub_list.insert(0, lbl_key)
            info.append(widgets.HBox(children=pub_list))
        elif key == 'experimenter':
            lbl_experimenter = widgets.Label('Experimenter:', layout=field_lay)
            if isinstance(value, (list, tuple)):
                lbl_names = widgets.Label(', '.join(value), layout=field_lay)
            else:
                lbl_names = widgets.Label(value, layout=field_lay)
            hbox_exp = widgets.HBox(children=[lbl_experimenter, lbl_names])
            info.append(hbox_exp)
        elif (isinstance(value, Iterable) and len(value)) or value:
            neuro_data.append(view.nwb2widget(value, neurodata_vis_spec=neurodata_vis_spec))
            labels.append(key)
    accordion = widgets.Accordion(children=neuro_data, selected_index=None)
    for i, label in enumerate(labels):
        if hasattr(node.fields[label], 'description') and node.fields[label].description:
            accordion.set_title(i, label + ': ' + node.fields[label].description)
        else:
            accordion.set_title(i, label)
    return widgets.VBox(info + [accordion])


def dict2accordion(d: dict, neurodata_vis_spec: dict, **pass_kwargs) -> widgets.Accordion:
    children = [widgets.HTML('Rendering...') for _ in d]
    accordion = widgets.Accordion(children=children, selected_index=None)
    for i, label in enumerate(d):
        if hasattr(d[label], 'description') and d[label].description:
            accordion.set_title(i, label + ': ' + d[label].description)
        else:
            accordion.set_title(i, label)
        accordion.set_title(i, label)

    def on_selected_index(change):
        if change.new is not None and isinstance(change.owner.children[change.new], widgets.HTML):
            children[change.new] = nwb2widget(list(d.values())[change.new], neurodata_vis_spec=neurodata_vis_spec,
                                              **pass_kwargs)
            change.owner.children = children

    accordion.observe(on_selected_index, names='selected_index')

    return accordion


def lazy_tabs(in_dict: dict, node, style: GroupingWidget = widgets.Tab) -> GroupingWidget:
    """Creates a lazy tab object where multiple visualizations can be used for a single node and are generated on the
    fly

    Parameters
    ----------
    in_dict: dict
        keys are labels for tabs and values are functions
    node: NWBDataInterface
        instance of neurodata type to visualize
    style: ipywidgets.Tab or ipywidgets.Accordion, optional
        which way to present the data

    Returns
    -------
    tab: widget

    """
    tabs_spec = list(in_dict.items())

    children = [tabs_spec[0][1](node)] + [widgets.HTML('Rendering...')
                                          for _ in range(len(tabs_spec) - 1)]
    tab = style(children=children)
    [tab.set_title(i, label) for i, (label, _) in enumerate(tabs_spec)]

    def on_selected_index(change):
        if isinstance(change.owner.children[change.new], widgets.HTML):
            children[change.new] = vis2widget(tabs_spec[change.new][1](node))
            change.owner.children = children

    tab.observe(on_selected_index, names='selected_index')

    return tab


def lazy_show_over_data(list_, func_, labels=None, style: GroupingWidget = widgets.Tab) -> GroupingWidget:
    """
    Apply same function to list of data in lazy tabs or lazy accordion
    Parameters
    ----------
    list_
    func_
    labels: list of str
    style: widgets.Tab or widgets.Accordion

    Returns
    -------
    ipywidgets.Tab or ipywidgets.Accordion
        subtype Tab or Accordion

    """
    children = [vis2widget(func_(list_[0]))] + [widgets.HTML('Rendering...') for _ in range(len(list_) - 1)]
    out = style(children=children)
    if labels is not None:
        [out.set_title(i, label) for i, label in enumerate(labels)]

    def on_selected_index(change):
        if change.new is not None and isinstance(change.owner.children[change.new], widgets.HTML):
            children[change.new] = vis2widget(func_(list_[change.new]))
            change.owner.children = children

    out.observe(on_selected_index, names='selected_index')

    return out


def nwb2widget(node,  neurodata_vis_spec: dict, **pass_kwargs) -> widgets.Widget:
    for ndtype in type(node).__mro__:
        if ndtype in neurodata_vis_spec:
            spec = neurodata_vis_spec[ndtype]
            if isinstance(spec, dict):
                return lazy_tabs(spec, node)
            elif callable(spec):
                return vis2widget(spec(node, neurodata_vis_spec=neurodata_vis_spec, **pass_kwargs))
    out1 = widgets.Output()
    with out1:
        print(node)
    return out1


def vis2widget(vis) -> widgets.Widget:
    if isinstance(vis, widgets.Widget):
        return vis
    elif isinstance(vis, plt.Figure):
        return fig2widget(vis)
    elif isinstance(vis, plt.Axes):
        return fig2widget(vis.get_figure())
    else:
        raise ValueError('unsupported vis type {}'.format(type(vis)))


def fig2widget(fig: Figure, **kwargs) -> widgets.Widget:
    out = widgets.Output()
    with out:
        plt.show(fig)
    return out


def processing_module(node: ProcessingModule, neurodata_vis_spec: dict) -> widgets.Widget:
    return nwb2widget(node.data_interfaces, neurodata_vis_spec=neurodata_vis_spec)


def show_text_fields(node, exclude=('comments', 'interval'), **kwargs) -> widgets.Widget:
    info = []
    for key in node.fields:
        if key not in exclude and isinstance(key, (str, float, int)):
            info.append(widgets.Text(value=repr(getattr(node, key)), description=key, disabled=True))
    return widgets.VBox(info)


def df2accordion(df: pd.DataFrame, by, func, style: GroupingWidget = widgets.Accordion, detect_single=True) \
        -> GroupingWidget:
    """
    Visualize pandas.DataFrame with an ipywidgets.Accordion

    Parameters
    ----------
    df: pandas.DataFrame
    by: str
    func: visualization function
    style: ipywidgets.Tab or ipywidgets.Accordion, optional
    detect_single: bool
        If True, test if the dimension you are grouping by only has 1 unique value. If so, do not form an Accordion.

    Returns
    -------
    ipywigets.Accordion or ipywidgets.Tab


    """
    if detect_single and df[by].nunique() == 1:
        return func(df)
    else:
        labels, idfs = zip(*df.groupby(by))
        return lazy_show_over_data(idfs, func, labels=labels, style=style)



from nwbwidgets.utils.cmaps import linear_transfer_function,matplotlib_transfer_function
from ipyvolume import TransferFunction

def test_linear_transfer_function():
    assert isinstance(linear_transfer_function('blue',reverse_opacity=True),TransferFunction)
    
def test_matplotlib_transfer_function():
    assert isinstance(matplotlib_transfer_function('PuBuGn_r'),TransferFunction)



from .base import lazy_show_over_data, GroupingWidget
from .timeseries import show_timeseries_mpl
from ipywidgets import widgets
import matplotlib.pyplot as plt
from ndx_icephys_meta.icephys import SweepSequences
from functools import partial
import numpy as np
from matplotlib.pyplot import Figure
import pandas as pd


def show_single_sweep_sequence(sweep_sequence, axs=None, title=None, **kwargs) -> Figure:
    """
    Show a single rep of a single stimulus sequence

    Parameters
    ----------
    sweep_sequence
    axs: [matplotlib.pyplot.Axes, matplotlib.pyplot.Axes], optional
    title: str, optional
    kwargs: dict
        passed to show_timeseries_mpl

    Returns
    -------

    matplotlib.pyplot.Figure

    """

    nsweeps = len(sweep_sequence)
    if axs is None:
        fig, axs = plt.subplots(2, 1, sharex=True)
    else:
        fig = axs[0].get_figure()
    for i in range(nsweeps):
        start, stop, ts = sweep_sequence['recordings'].iloc[i]['response'].iloc[0]
        show_timeseries_mpl(ts, istart=start, istop=stop, ax=axs[0], zero_start=True, xlabel='', title=title, **kwargs)

        start, stop, ts = sweep_sequence['recordings'].iloc[i]['stimulus'].iloc[0]
        show_timeseries_mpl(ts, istart=start, istop=stop, ax=axs[1], zero_start=True, **kwargs)
    return fig


def show_sweep_sequence_reps(stim_df: pd.DataFrame, **kwargs) -> Figure:
    """
    Show data from multiple reps of the same stimulus type

    Parameters
    ----------
    stim_df: pandas.DataFrame
    kwargs: dict
        passed to show_single_sweep_sequence

    Returns
    -------
    matplotlib.pyplot.Figure

    """
    nsweeps = len(stim_df['sweeps'])

    if 'repetition' in stim_df:
        stim_df = stim_df.sort_values('repetition')
    fig, axs = plt.subplots(2, nsweeps, sharex='col', sharey='row', figsize=[6.4 * nsweeps, 4.8])
    if nsweeps == 1:
        axs = np.array([axs]).T
    for i, (sweep, sweep_axs) in enumerate(zip(stim_df['sweeps'], axs.T)):
        if i:
            kwargs.update(ylabel='')
        show_single_sweep_sequence(sweep, axs=sweep_axs, title='rep {}'.format(i+1), **kwargs)
    return fig


def show_sweep_sequences(node: SweepSequences, *args, style: GroupingWidget = widgets.Accordion, **kwargs) -> \
        GroupingWidget:
    """
    Visualize the sweep sequences table with a lazy accordion of sweep sequence repetitions

    Parameters
    ----------
    node: SweepSequences
    style: widgets.Accordion or widgets.Tabs

    Returns
    -------
    widgets.Accordion or widgets.Tabs

    """
    if 'stimulus_type' in node:
        labels, data = zip(*[(stim_label, stim_df)
                             for stim_label, stim_df in node.to_dataframe().groupby('stimulus_type')])
        func_ = show_sweep_sequence_reps
    else:
        data = node['sweeps']
        labels = None
        func_ = show_single_sweep_sequence
    func_ = partial(func_, **kwargs)
    return lazy_show_over_data(data, func_, labels=labels, style=style)



import pynwb
import ndx_grayscalevolume
from collections import OrderedDict
from nwbwidgets import behavior, misc, base, ecephys, image, ophys, icephys, timeseries, file
import hdmf
from functools import partial
from ndx_icephys_meta.icephys import SweepSequences


default_neurodata_vis_spec = {
    pynwb.NWBFile: file.show_nwbfile,
    SweepSequences: icephys.show_sweep_sequences,
    pynwb.behavior.BehavioralEvents: behavior.show_behavioral_events,
    pynwb.ecephys.LFP: ecephys.show_lfp,
    pynwb.misc.Units: OrderedDict({
        'Session Raster': misc.RasterWidget,
        'Grouped PSTH': misc.PSTHWidget,
        'Raster Grid': misc.RasterGridWidget,
        'table': base.show_dynamic_table}),
    pynwb.misc.DecompositionSeries: misc.show_decomposition_series,
    pynwb.file.Subject: base.show_fields,
    pynwb.ophys.ImagingPlane: base.show_fields,
    pynwb.ecephys.SpikeEventSeries: ecephys.show_spike_event_series,
    pynwb.ophys.ImageSegmentation: ophys.show_image_segmentation,
    pynwb.ophys.TwoPhotonSeries: ophys.show_two_photon_series,
    ndx_grayscalevolume.GrayscaleVolume: ophys.show_grayscale_volume,
    pynwb.ophys.PlaneSegmentation: ophys.show_plane_segmentation,
    pynwb.ophys.DfOverF: ophys.show_df_over_f,
    pynwb.ophys.RoiResponseSeries: timeseries.traces_widget,
    pynwb.misc.AnnotationSeries: OrderedDict({
        'text': base.show_text_fields,
        'times': misc.show_annotations}),
    pynwb.core.LabelledDict: base.dict2accordion,
    pynwb.ProcessingModule: base.processing_module,
    hdmf.common.DynamicTable: base.show_dynamic_table,
    pynwb.ecephys.ElectricalSeries: OrderedDict({
        'Fields': timeseries.show_ts_fields,
        'Traces': partial(timeseries.traces_widget,
                          start=0, dur=10,
                          trace_starting_range=(0, 5)),
    }),
    pynwb.behavior.Position: behavior.show_position,
    pynwb.behavior.SpatialSeries: OrderedDict({
        'over time': behavior.show_spatial_series_over_time,
        'trace': behavior.show_spatial_series}),
    pynwb.image.GrayscaleImage: image.show_grayscale_image,
    pynwb.image.RGBImage: image.show_rbg_image,
    pynwb.image.ImageSeries: image.show_image_series,
    pynwb.image.IndexSeries: image.show_index_series,
    pynwb.TimeSeries: timeseries.show_timeseries,
    pynwb.core.NWBDataInterface: base.show_neurodata_base,
}


def nwb2widget(node,  neurodata_vis_spec=default_neurodata_vis_spec):
    return base.nwb2widget(node, neurodata_vis_spec)



import numpy as np
import matplotlib.pyplot as plt
from ipywidgets import widgets, fixed
from pynwb import TimeSeries
import pynwb
from .utils.timeseries import (get_timeseries_tt, get_timeseries_maxt, get_timeseries_mint,
                               timeseries_time_to_ind, get_timeseries_in_units)
from .controllers import StartAndDurationController,  RangeController
from .base import fig2widget


def show_ts_fields(node):
    info = []
    for key in ('description', 'unit', 'resolution', 'conversion'):
        info.append(widgets.Text(value=repr(getattr(node, key)), description=key, disabled=True))
    return widgets.VBox(info)


def show_timeseries_mpl(node: TimeSeries, neurodata_vis_spec=None, istart=0, istop=None, ax=None, zero_start=False,
                        xlabel=None, ylabel=None, title=None, **kwargs):
    if xlabel is None:
        xlabel = 'time (s)'

    if ylabel is None and node.unit:
        ylabel = node.unit

    if ax is None:
        fig, ax = plt.subplots()

    tt = get_timeseries_tt(node, istart=istart, istop=istop)
    if zero_start:
        tt = tt - tt[0]
    data, unit = get_timeseries_in_units(node, istart=istart, istop=istop)

    ax.plot(tt, data, **kwargs)
    ax.set_xlabel(xlabel)
    if node.unit:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.autoscale(enable=True, axis='x', tight=True)

    return ax


def show_timeseries(node: TimeSeries, neurodata_vis_spec=None, istart=0, istop=None, ax=None, zero_start=False,
                    **kwargs):
    fig = show_timeseries_mpl(node, neurodata_vis_spec, istart, istop, ax, zero_start, **kwargs)

    info = []
    for key in ('description', 'comments', 'unit', 'resolution', 'conversion'):
        info.append(widgets.Text(value=repr(getattr(node, key)), description=key, disabled=True))
    children = [widgets.VBox(info)]

    children.append(fig2widget(fig))

    return widgets.HBox(children=children)


def plot_traces(time_series: TimeSeries, time_window=None, trace_window=None,
                title: str = None, ylabel: str = 'traces'):
    """

    Parameters
    ----------
    time_series: pynwb.TimeSeries
    time_start: float
        Start time in seconds
    time_duration: float, optional
        Duration in seconds. Default:
    trace_window: [int int], optional
        Index range of traces to view
    title: str, optional
    ylabel: str, optional

    Returns
    -------

    """

    if time_window is None:
        t_ind_start = 0
        t_ind_stop = None
    else:
        t_ind_start = timeseries_time_to_ind(time_series, time_window[0])
        t_ind_stop = timeseries_time_to_ind(time_series, time_window[1])

    if trace_window is None:
        trace_window = [0, time_series.data.shape[1]]
    tt = get_timeseries_tt(time_series, t_ind_start, t_ind_stop)
    if time_series.data.shape[1] == len(tt):  # fix of orientation is incorrect
        mini_data = time_series.data[trace_window[0]:trace_window[1], t_ind_start:t_ind_stop].T
    else:
        mini_data = time_series.data[t_ind_start:t_ind_stop, trace_window[0]:trace_window[1]]

    gap = np.median(np.nanstd(mini_data, axis=0)) * 20
    offsets = np.arange(trace_window[1] - trace_window[0]) * gap

    fig, ax = plt.subplots()
    ax.figure.set_size_inches(12, 6)
    ax.plot(tt, mini_data + offsets)
    ax.set_xlabel('time (s)')
    if np.isfinite(gap):
        ax.set_ylim(-gap, offsets[-1] + gap)
        ax.set_xlim(tt[0], tt[-1])
        ax.set_yticks(offsets)
        ax.set_yticklabels(np.arange(trace_window[0], trace_window[1]))

    if title is not None:
        ax.set_title(title)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    return fig


def traces_widget(node: TimeSeries, neurodata_vis_spec: dict = None,
                  time_window_controller=None, start=None, dur=None,
                  trace_controller=None, trace_starting_range=None,
                  **kwargs):

    if time_window_controller is None:
        tmax = get_timeseries_maxt(node)
        tmin = get_timeseries_mint(node)
        if start is None:
            start = tmin
        if dur is None:
            dur = min(tmax-tmin, 5)
        time_window_controller = StartAndDurationController(tmax, tmin, start=start, duration=dur)
    if trace_controller is None:
        if trace_starting_range is None:
            trace_starting_range = (0, min(30, node.data.shape[1]))
        trace_controller = RangeController(0, node.data.shape[1], start_range=trace_starting_range,
                                           description='channels', dtype='int', orientation='vertical')

    controls = {
        'time_series': widgets.fixed(node),
        'time_window': time_window_controller,
        'trace_window': trace_controller.slider,
    }
    controls.update({key: widgets.fixed(val) for key, val in kwargs.items()})

    out_fig = widgets.interactive_output(plot_traces, controls)

    lower = widgets.HBox(children=[
        trace_controller,
        out_fig
    ])

    out = widgets.VBox(children=[
        time_window_controller,
        lower
    ])

    return out


def single_trace_widget(timeseries: TimeSeries, time_window_controller=None):

    controls = dict(timeseries=fixed(timeseries))

    gen_time_window_controller = False
    if time_window_controller is None:
        gen_time_window_controller = True
        tmin = get_timeseries_mint(timeseries)
        tmax = get_timeseries_maxt(timeseries)
        time_window_controller = RangeController(tmin, tmax, start_value=[tmin, min(tmin+30, tmax)])

    controls.update(time_window=time_window_controller.slider)

    out_fig = widgets.interactive_output(show_trace, controls)

    if gen_time_window_controller:
        return widgets.VBox(children=[time_window_controller, out_fig])
    else:
        return widgets.VBox(children=[out_fig])


def show_trace(timeseries, time_window):
    istart = timeseries_time_to_ind(timeseries, time_window[0])
    istop = timeseries_time_to_ind(timeseries, time_window[1])

    return show_timeseries_mpl(timeseries, istart=istart, istop=istop).get_figure()


from pynwb import TimeSeries

import numpy as np
from bisect import bisect


def get_timeseries_tt(node: TimeSeries, istart=0, istop=None) -> np.ndarray:
    """
    For any TimeSeries, return timestamps. If the TimeSeries uses starting_time and rate, the timestamps will be
    generated.

    Parameters
    ----------
    node: pynwb.TimeSeries
    istart: int, optional
        Optionally sub-select the returned times - lower bound
    istop: int, optional
        Optionally sub-select the returned times - upper bound

    Returns
    -------
    numpy.ndarray

    """
    if node.timestamps is not None:
        return node.timestamps[istart:istop]
    else:
        if not np.isfinite(node.starting_time):
            starting_time = 0
        else:
            starting_time = node.starting_time
        if istop is None:
            return np.arange(istart, len(node.data)) / node.rate + starting_time
        elif istop > 0:
            return np.arange(istart, istop) / node.rate + starting_time
        else:
            return np.arange(istart, len(node.data) + istop - 1) / node.rate + starting_time


def get_timeseries_maxt(node: TimeSeries) -> float:
    """
    Returns the maximum time of any TimeSeries

    Parameters
    ----------
    node: pynwb.TimeSeries

    Returns
    -------
    float

    """
    if node.timestamps is not None:
        return node.timestamps[-1]
    else:
        return len(node.data) / node.rate + node.starting_time


def get_timeseries_mint(node: TimeSeries) -> float:
    """
    Returns the minimum time of any TimeSeries

    Parameters
    ----------
    node: pynwb.TimeSeries

    Returns
    -------
    float

    """
    if node.timestamps is not None:
        return node.timestamps[0]
    else:
        return node.starting_time


def get_timeseries_in_units(node: TimeSeries, istart=None, istop=None):
    """
    Convert data into the designated units

    Parameters
    ----------
    node: pynwb.TimeSeries
    istart: int
    istop: int

    Returns
    -------
    numpy.ndarray, str

    """
    data = node.data[istart:istop]
    if node.conversion and np.isfinite(node.conversion):
        data = data * node.conversion
        unit = node.unit
    else:
        unit = None
    return data, unit


def timeseries_time_to_ind(node: TimeSeries, time, ind_min=None, ind_max=None) -> int:
    """
    Get the index of a certain time for any TimeSeries. For TimeSeries that use timestamps, bisect is used. You can
    optionally provide ind_min and ind_max to constrain the search.

    Parameters
    ----------
    node: pynwb.TimeSeries
    time: float
    ind_min: int, optional
    ind_max: int, optional

    Returns
    -------

    """
    if node.timestamps is not None:
        kwargs = dict()
        if ind_min is not None:
            kwargs.update(lo=ind_min)
        if ind_max is not None:
            kwargs.update(hi=ind_max)
        return bisect(node.timestamps, time, **kwargs)
    else:
        return int(np.ceil((time - node.starting_time) * node.rate))


def align_by_times(timeseries: TimeSeries, starts, stops):
    """
    Args:
        timeseries: TimeSeries
        starts: array-like
        stops: array-like
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    out = []
    for istart, istop in zip(starts, stops):
        ind_start = timeseries_time_to_ind(timeseries, istart)
        ind_stop = timeseries_time_to_ind(timeseries, istop, ind_min=ind_start)
        out.append(timeseries.data[ind_start:ind_stop])
    return np.array(out)


def align_by_trials(timeseries: TimeSeries, start_label='start_time',
                    stop_label=None, before=0., after=1.):
    """
    Args:
        timeseries: TimeSeries
        start_label: str
            default: 'start_time'
        stop_label: str
            default: None (just align to start_time)
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    trials = timeseries.get_ancestor('NWBFile').trials
    return align_by_time_intervals(timeseries, trials, start_label, stop_label, before, after)


def align_by_time_intervals(timeseries: TimeSeries, intervals, start_label='start_time',
                            stop_label='stop_time', before=0., after=0.):
    """
    Args:
        timeseries: pynwb.TimeSeries
        intervals: pynwb.epoch.TimeIntervals
        start_label: str
            default: 'start_time'
        stop_label: str
            default: 'stop_time'
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    if stop_label is None:
        stop_label = 'start_time'

    starts = np.array(intervals[start_label][:]) - before
    stops = np.array(intervals[stop_label][:]) + after
    return align_by_times(timeseries, starts, stops)



import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
import ipywidgets as widgets
from ndx_grayscalevolume import GrayscaleVolume
from nwbwidgets.view import default_neurodata_vis_spec
from pynwb.ophys import TwoPhotonSeries, OpticalChannel, ImageSegmentation, Fluorescence, DfOverF
from pynwb.device import Device
from nwbwidgets.ophys import show_grayscale_volume,show_two_photon_series,show_df_over_f
import unittest



def test_show_grayscale_volume():
    vol = GrayscaleVolume(name='vol',data=np.random.rand(2700).reshape((30,30,3)))
    assert isinstance(show_grayscale_volume(vol, default_neurodata_vis_spec),widgets.Widget)
    
    
class CalciumImagingTestCase(unittest.TestCase):
    
    def setUp(self):
        
        nwbfile = NWBFile('my first synthetic recording', 'EXAMPLE_ID', datetime.now(tzlocal()),
                          experimenter='Dr. Bilbo Baggins',
                          lab='Bag End Laboratory',
                          institution='University of Middle Earth at the Shire',
                          experiment_description=('I went on an adventure with thirteen '
                                                  'dwarves to reclaim vast treasures.'),
                          session_id='LONELYMTN')


        device = Device('imaging_device_1')
        nwbfile.add_device(device)
        optical_channel = OpticalChannel('my_optchan', 'description', 500.)
        imaging_plane = nwbfile.create_imaging_plane('my_imgpln', optical_channel, 'a very interesting part of the brain',
                                                     device, 600., 300., 'GFP', 'my favorite brain location',
                                                     np.ones((5, 5, 3)), 4.0, 'manifold unit', 'A frame to refer to')

        self.image_series = TwoPhotonSeries(name='test_iS', dimension=[2],data=np.random.rand(10, 5, 5, 3),
                                       external_file=['images.tiff'], imaging_plane=imaging_plane,
                                       starting_frame=[0], format='tiff', starting_time=0.0, rate=1.0)
        nwbfile.add_acquisition(self.image_series)


        mod = nwbfile.create_processing_module('ophys', 'contains optical physiology processed data')
        img_seg = ImageSegmentation()
        mod.add(img_seg)
        ps = img_seg.create_plane_segmentation('output from segmenting my favorite imaging plane',
                                               imaging_plane, 'my_planeseg', self.image_series)


        w, h = 3, 3
        pix_mask1 = [(0, 0, 1.1), (1, 1, 1.2), (2, 2, 1.3)]
        vox_mask1 = [(0, 0, 0, 1.1), (1, 1, 1, 1.2), (2, 2, 2, 1.3)]
        img_mask1 = [[0.0 for x in range(w)] for y in range(h)]
        img_mask1[0][0] = 1.1
        img_mask1[1][1] = 1.2
        img_mask1[2][2] = 1.3
        ps.add_roi(pixel_mask=pix_mask1, image_mask=img_mask1, voxel_mask=vox_mask1)

        pix_mask2 = [(0, 0, 2.1), (1, 1, 2.2)]
        vox_mask2 = [(0, 0, 0, 2.1), (1, 1, 1, 2.2)]
        img_mask2 = [[0.0 for x in range(w)] for y in range(h)]
        img_mask2[0][0] = 2.1
        img_mask2[1][1] = 2.2
        ps.add_roi(pixel_mask=pix_mask2, image_mask=img_mask2, voxel_mask=vox_mask2)

        fl = Fluorescence()
        mod.add(fl)

        rt_region = ps.create_roi_table_region('the first of two ROIs', region=[0])

        data = np.random.randn(10, 5)
        timestamps = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
        rrs = fl.create_roi_response_series('my_rrs', data, rt_region, unit='lumens', timestamps=timestamps)

        self.df_over_f = DfOverF(rrs)
        
    def test_show_two_photon_series(self):
        assert isinstance(show_two_photon_series(self.image_series, default_neurodata_vis_spec), widgets.Widget)
        
    def test_show_df_over_f(self):
        assert isinstance(show_df_over_f(self.df_over_f, default_neurodata_vis_spec), widgets.Widget)




import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
from ipywidgets import widgets
from pynwb.epoch import TimeIntervals
from nwbwidgets.utils.units import get_min_spike_time, align_by_trials, align_by_time_intervals
import unittest

class ShowPSTHTestCase(unittest.TestCase):

    def setUp(self):

        start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
        create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

        self.nwbfile = NWBFile(session_description='NWBFile for PSTH',
                               identifier='NWB123',
                               session_start_time=start_time,
                               file_create_date=create_date)
        
        self.nwbfile.add_unit_column('location', 'the anatomical location of this unit')
        self.nwbfile.add_unit_column('quality', 'the quality for the inference of this unit')
        
        self.nwbfile.add_unit(id=1, spike_times=[2.2, 3.0, 4.5],
                              obs_intervals=[[1, 10]], location='CA1', quality=0.95)
        self.nwbfile.add_unit(id=2, spike_times=[2.2, 3.0, 25.0, 26.0],
                              obs_intervals=[[1, 10], [20, 30]], location='CA3', quality=0.85)
        self.nwbfile.add_unit(id=3, spike_times=[1.2, 2.3, 3.3, 4.5],
                              obs_intervals=[[1, 10], [20, 30]], location='CA1', quality=0.90)
        
        self.nwbfile.add_trial_column(name='stim', description='the visual stimuli during the trial')

        self.nwbfile.add_trial(start_time=0.0, stop_time=2.0, stim='person')
        self.nwbfile.add_trial(start_time=3.0, stop_time=5.0, stim='ocean')
        self.nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim='desert')

    def test_get_min_spike_time(self):
        assert(get_min_spike_time(self.nwbfile.units)==1.2)
        
    def test_align_by_trials(self):
        ComparetoAT = [np.array([ 2.2,  3.0 , 25.0 , 26.0 ]),np.array([-0.8,  0. , 22. , 23. ]),
                       np.array([-3.8, -3. , 19. , 20. ])]
        
        AT = align_by_trials(self.nwbfile.units, index=1, before=20., after=30.)
        
        np.testing.assert_allclose(AT, ComparetoAT, rtol=1e-02)
        
    def test_align_by_time_intervals_Nonetrials_select(self):
        time_intervals = TimeIntervals(name='Test Time Interval')
        time_intervals.add_interval(start_time=21.0, stop_time=28.0)
        time_intervals.add_interval(start_time=22.0, stop_time=26.0)
        time_intervals.add_interval(start_time=22.0, stop_time=28.4)
        
        ATI = align_by_time_intervals(self.nwbfile.units, index=1, intervals=time_intervals,
                                      stop_label=None, before=20., after=30.)

        ComparedtoATI = [np.array([-18.8, -18. ,   4. ,   5. ]), np.array([-19.8, -19. ,   3. ,   4. ]), np.array([-19.8, -19. ,   3. ,   4. ])]
        
        np.testing.assert_array_equal(ATI, ComparedtoATI)
        
    def test_align_by_time_intervals(self):
        time_intervals = TimeIntervals(name='Test Time Interval')
        time_intervals.add_interval(start_time=21.0, stop_time=28.0)
        time_intervals.add_interval(start_time=22.0, stop_time=26.0)
        time_intervals.add_interval(start_time=22.0, stop_time=28.4)
        
        ATI = align_by_time_intervals(self.nwbfile.units, index=1, intervals=time_intervals,
                                      stop_label=None, before=20., after=30., rows_select=[0, 1])
        
        ComparedtoATI = [np.array([-18.8, -18., 4., 5.]), np.array([-19.8, -19., 3., 4.])]
        
        np.testing.assert_array_equal(ATI, ComparedtoATI)
        
        






import numpy as np
import scipy


def compute_smoothed_firing_rate(spike_times, tt, sigma_in_secs):
    """ Evaluate gaussian smoothing of spike_times at uniformly spaced array t
        Args:
          spike_times:
              A 1D numpy ndarray of spike times
          tt:
              1D array, uniformly spaced, e.g. the output of np.linspace or np.arange
          sigma_in_secs:
              standard deviation of the smoothing gaussian in seconds
        Returns:
              Gaussian smoothing evaluated at array t
        """
    binned_spikes = np.zeros_like(tt)
    binned_spikes[np.searchsorted(tt, spike_times)] += 1
    dt = np.diff(tt[:2])[0]
    sigma_in_samps = sigma_in_secs / dt
    smooth_fr = scipy.ndimage.gaussian_filter1d(binned_spikes, sigma_in_samps) / dt
    return smooth_fr


# ported from the chronux MATLAB package
def psth(data=None, sig=0.05, T=None, err=2, t=None, num_bootstraps=10):
    """ Find peristimulus time histogram smoothed by a gaussian kernel
        The time units of the arrays in data, sig and t
        should be the same, e.g. seconds
    Args:
      data:
            A dictionary of channel names, and 1D numpy ndarray spike times,
            A numpy ndarray, where each row gives the spike times for each channel
            A 1D numpy ndarray, one list or tuple of floats that gives spike times for only one channel
      sig:  standard deviation of the smoothing gaussian. default 0.05
      T: time interval [a,b], spike times strictly outside this interval are excluded
      err: An integer, 0, 1, or 2. default 2
            0 indicates no standard error computation
            1 Poisson error
            2 Boostrap method over trials
      t: 1D array, list or tuple indicating times to evaluate psth at
      num_bootstraps: number of bootstraps. Effective only in computing error when err=2. default 10
    Returns:
      R: Rate, mean smoothed peristimulus time histogram
      t: 1D array, list or tuple indicating times psth is evaluated at
      E: standard error
    """

    # verify data argument
    try:
        if isinstance(data, dict):
            data = data.values()
        elif isinstance(data[0], (float, int)):
            data = np.array([data])
        elif isinstance(data[0], (np.ndarray, list, tuple)):
            data = np.array(data)
        else:
            raise TypeError
        if isinstance(data, np.ndarray) and len(data.shape) == 2:
            if data.shape[0] == 0 or data.shape[1] == 0:
                raise TypeError
        else:
            data = [np.array(ch_data) for ch_data in data]
            for ch_data in data:
                if len(ch_data.shape) != 1 or not isinstance(ch_data[0], (float, int)):
                    raise TypeError
    except Exception as exc:
        msg = ("psth requires spike time data as first positional argument. " +
               "Spike time data should be in the form of:\n" +
               "   a dictionary of channel names, and 1D numpy ndarray spike times\n" +
               "   a 2D numpy ndarray, where each row represents the spike times of a channel", )
        exc.args = msg
        raise exc

    # input data size
    num_channels = len(data)
    channel_lengths = [len(ch_data) for ch_data in data]
    max_channel_length = max(channel_lengths)

    if not isinstance(sig, (float, int)) or sig <= 0:
        raise TypeError("sig must be positive. Only the non-adaptive method is supported")
    if not isinstance(num_bootstraps, int) or num_bootstraps <= 0:
        raise TypeError("num_bootstraps must be a positive integer")

    # determine the interval of interest T, and mask times outside of the interval
    if T is not None:
        # expand T to avoid edge effects in rate
        T = [T[0]-4*sig, T[1]+4*sig]
        data = [np.ma.masked_outside(np.ravel(ch_data), T[0], T[1]) for ch_data in data]
    else:
        T = [np.ma.min([np.ma.min(c) for c in data]), np.ma.max([np.ma.max(c) for c in data])]

    # determine t
    if t is None:
        num_points = int(5*(T[1]-T[0])/sig)
        t = np.linspace(T[0], T[1], num_points)
        t_min, t_max = T[0], T[1]
    else:
        t = np.ravel(t)
        num_points = len(t)
        t_min, t_max = np.min(t), np.max(t)

    # masked input data size
    data_lengths = [np.ma.count(ch_data) for ch_data in data]
    num_times_total = sum(data_lengths)+1

    # Transposes data if necessary. The longer dimension is assumed to be time.
    # This is to convert the data into a form compatible with Murray's routine
    if num_channels < max_channel_length:
        num_t = num_channels  # number of trials
    else:
        num_t = max_channel_length
        # pad and transpose
        data_transposed = np.ma.zeros((max_channel_length, num_channels))
        data_transposed[...] = np.ma.masked
        for n in range(num_channels):
            data_transposed[:, n] = data[n]
        data = data_transposed
        del data_transposed
        max_channel_length = num_channels

    # warn if spikes have low density
    L = num_times_total/(num_t*(T[1]-T[0]))
    if 2*L*num_t*sig < 1 or L < 0.1:
        print('Spikes have very low density. The time units may not be the same, or the kernel width is too small')
        print('Total events: %f \nsig: %f ms \nT: %f \nevents*sig: %f\n'
              % (num_times_total, sig*1000, T, num_times_total*sig/(T[1]-T[0])))

    # fine grid in case t does not have sufficient precision for gaussian_filter1d
    num_points_extended = 6*int(5*(t_max-t_min)/sig)
    t_extended = np.linspace(t_min, t_max, num_points_extended)
    smooth_fr_index = np.rint((t-t_min)/(t_extended[1]-t_extended[0])).astype(np.int)
    # evaluate kernel density estimation at array t
    RR = np.zeros((num_t, num_points))
    for n in range(num_t):
        spike_times = data[n] if not np.ma.is_masked(data[n]) else data[n].compressed()
        smooth_fr = compute_smoothed_firing_rate(spike_times, t_extended, sig)
        RR[n, :] = smooth_fr[smooth_fr_index]

    # find rate
    R = np.mean(RR, axis=0)

    # find error
    if num_t < 4 and err == 2:
        print('Switching to Poisson errorbars as number of trials is too small for bootstrap')
        err = 1
    # std dev is sqrt(rate*(integral over kernal^2)/trials)
    # for Gaussian integral over Kernal^2 is 1/(2*sig*srqt(pi))
    if err == 0:
        E = None
    elif err == 1:
        E = np.sqrt(R/(2*num_t*sig*np.sqrt(np.pi)))
    elif err == 2:
        mean_ = [np.mean(RR[np.random.randint(0, num_t), :]) for _ in range(num_bootstraps)]
        E = np.std(mean_)
    else:
        raise TypeError("err must be 0, 1, or 2")

    return R, t, E


import numpy as np
import matplotlib.pyplot as plt
from pynwb.ophys import RoiResponseSeries, DfOverF, PlaneSegmentation, TwoPhotonSeries, ImageSegmentation
from pynwb.base import NWBDataInterface
from ndx_grayscalevolume import GrayscaleVolume
from .utils.cmaps import linear_transfer_function
from .utils.dynamictable import infer_categorical_columns
from .utils.functional import MemoizeMutable
import ipywidgets as widgets
import plotly.graph_objects as go
from skimage import measure


color_wheel = ['red', 'blue', 'green', 'black', 'magenta', 'yellow']


def show_two_photon_series(indexed_timeseries: TwoPhotonSeries, neurodata_vis_spec: dict):
    output = widgets.Output()

    if len(indexed_timeseries.data.shape) == 3:
        def show_image(index=0):
            fig, ax = plt.subplots(subplot_kw={'xticks': [], 'yticks': []})
            ax.imshow(indexed_timeseries.data[index], cmap='gray')
            output.clear_output(wait=True)
            with output:
                plt.show(fig)
    elif len(indexed_timeseries.data.shape) == 4:
        import ipyvolume.pylab as p3

        def show_image(index=0):
            p3.figure()
            p3.volshow(indexed_timeseries.data[index], tf=linear_transfer_function([0, 0, 0], max_opacity=.3))
            output.clear_output(wait=True)
            with output:
                p3.show()
    else:
        raise NotImplementedError

    slider = widgets.IntSlider(value=0, min=0,
                               max=indexed_timeseries.data.shape[0] - 1,
                               orientation='horizontal')
    slider.observe(lambda change: show_image(change.new), names='value')
    show_image()

    return widgets.VBox([output, slider])


def show_df_over_f(df_over_f: DfOverF, neurodata_vis_spec: dict):
    if len(df_over_f.roi_response_series) == 1:
        title, input = list(df_over_f.roi_response_series.items())[0]
        return neurodata_vis_spec[RoiResponseSeries](input, neurodata_vis_spec, title=title)
    else:
        return neurodata_vis_spec[NWBDataInterface](df_over_f, neurodata_vis_spec)


def show_image_segmentation(img_seg: ImageSegmentation, neurodata_vis_spec: dict):
    if len(img_seg.plane_segmentations) == 1:
        return show_plane_segmentation(list(img_seg.plane_segmentations.values())[0], neurodata_vis_spec)
    else:
        return neurodata_vis_spec[NWBDataInterface](img_seg, neurodata_vis_spec)


def show_plane_segmentation_3d(plane_seg: PlaneSegmentation):
    import ipyvolume.pylab as p3

    nrois = len(plane_seg)

    dims = np.array([max(max(plane_seg['voxel_mask'][i][dim]) for i in range(nrois))
                     for dim in ['x', 'y', 'z']]).astype('int') + 1
    fig = p3.figure()
    for icolor, color in enumerate(color_wheel):
        vol = np.zeros(dims)
        sel = np.arange(icolor, nrois, len(color_wheel))
        for isel in sel:
            dat = plane_seg['voxel_mask'][isel]
            vol[tuple(dat['x'].astype('int')),
                tuple(dat['y'].astype('int')),
                tuple(dat['z'].astype('int'))] = 1
        p3.volshow(vol, tf=linear_transfer_function(color, max_opacity=.3))
    return fig


def compute_outline(image_mask, threshold):
    x, y = zip(*measure.find_contours(image_mask, threshold)[0])
    return x, y


compute_outline = MemoizeMutable(compute_outline)


def show_plane_segmentation_2d(plane_seg: PlaneSegmentation, color_wheel=color_wheel, color_by='neuron_type',
                               threshold=.01, fig=None):
    """

    Parameters
    ----------
    plane_seg
    color_wheel
    color_by
    threshold
    fig: plotly.graph_objects.Figure, options

    Returns
    -------

    """
    if color_by:
        if color_by in plane_seg:
            cats = np.unique(plane_seg[color_by][:])
        else:
            raise ValueError('specified color_by parameter, {}, not in plane_seg object'.format(color_by))
    data = plane_seg['image_mask'].data
    nUnits = data.shape[0]
    if fig is None:
        fig = go.FigureWidget()
    else:
        fig.data = None
    aux_leg = []
    for i in range(nUnits):
        if plane_seg[color_by][i] not in aux_leg:
            show_leg = True
            aux_leg.append(plane_seg[color_by][i])
        else:
            show_leg = False
        kwargs = dict()
        if color_by:
            c = color_wheel[np.where(cats == plane_seg[color_by][i])[0][0]]
            kwargs.update(line_color=c)
        # hover text
        hovertext = '<b>roi_id</b>: ' + str(plane_seg.id[i])
        rois_cols = list(plane_seg.colnames)
        if 'roi_id' in rois_cols:
            rois_cols.remove('roi_id')
        sec_str = '<br>'.join([col + ': ' + str(plane_seg[col][i]) for col in rois_cols if
                               isinstance(plane_seg[col][i], (int, float, np.integer, np.float, str))])
        hovertext += '<br>' + sec_str
        # form cell borders
        x, y = compute_outline(plane_seg['image_mask'][i], threshold)

        fig.add_trace(
            go.Scatter(
                x=x, y=y,
                fill='toself',
                mode='lines',
                name=str(plane_seg[color_by][i]),
                legendgroup=str(plane_seg[color_by][i]),
                showlegend=show_leg,
                text=hovertext,
                hovertext='text',
                line=dict(width=.5),
                **kwargs
            )
        )
        fig.update_layout(
            width=700, height=500,
            margin=go.layout.Margin(l=60, r=60, b=60, t=60, pad=1),
            plot_bgcolor="rgb(245, 245, 245)",
        )
    return fig


def plane_segmentation_2d_widget(plane_seg: PlaneSegmentation, **kwargs):

    categorical_columns = infer_categorical_columns(plane_seg)

    if len(categorical_columns) == 1:
        color_by = list(categorical_columns.keys())[0]
        return show_plane_segmentation_2d(plane_seg, color_by=color_by, **kwargs)

    elif len(categorical_columns) > 1:
        cat_controller = widgets.Dropdown(options=list(categorical_columns), description='color by')

        out_fig = show_plane_segmentation_2d(plane_seg, color_by=cat_controller.value, **kwargs)

        def on_change(change, out_fig=out_fig):
            if change['new'] and isinstance(change['new'], dict):
                ind = change['new']['index']
                if isinstance(ind, int):
                    color_by = change['owner'].options[ind]
                    show_plane_segmentation_2d(plane_seg, color_by=color_by, fig=out_fig, **kwargs)

        cat_controller.observe(on_change)

        return widgets.VBox(children=[cat_controller, out_fig])
    else:
        return show_plane_segmentation_2d(plane_seg, color_by=None, **kwargs)


def show_plane_segmentation(plane_seg: PlaneSegmentation, neurodata_vis_spec: dict):
    if 'voxel_mask' in plane_seg:
        return show_plane_segmentation_3d(plane_seg)
    elif 'image_mask' in plane_seg:
        return plane_segmentation_2d_widget(plane_seg)


def show_grayscale_volume(vol: GrayscaleVolume, neurodata_vis_spec: dict):
    import ipyvolume.pylab as p3

    fig = p3.figure()
    p3.volshow(vol.data, tf=linear_transfer_function([0, 0, 0], max_opacity=.1))
    return fig




import matplotlib.pyplot as plt
import numpy as np
import pynwb
import scipy
from ipywidgets import widgets, fixed, FloatProgress
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
from pynwb.misc import AnnotationSeries, Units, DecompositionSeries

from .controllers import make_trial_event_controller, GroupAndSortController, StartAndDurationController, ProgressBar
from .utils.dynamictable import infer_categorical_columns
from .utils.units import get_spike_times, get_max_spike_time, get_min_spike_time, align_by_time_intervals, \
    get_unobserved_intervals
from .utils.mpl import create_big_ax
from .utils.widgets import interactive_output
from .analysis.spikes import compute_smoothed_firing_rate

from ipywidgets import Layout

color_wheel = plt.rcParams['axes.prop_cycle'].by_key()['color']


def show_annotations(annotations: AnnotationSeries, **kwargs):
    fig, ax = plt.subplots()
    ax.eventplot(annotations.timestamps, **kwargs)
    ax.set_xlabel('time (s)')
    return fig


def show_session_raster(units: Units, time_window=None, units_window=None, show_obs_intervals=True,
                        order=None, group_inds=None, labels=None, show_legend=True, progress_bar=None):
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    time_window: [int, int]
    units_window: [int, int]
    show_obs_intervals: bool
    order: array-like, optional
    group_inds: array-like, optional
    labels: array-like, optional
    show_legend: bool
        default = True
        Does not show legend if color_by is None or 'id'.
    progress_bar: FloatProgress, optional

    Returns
    -------
    matplotlib.pyplot.Figure

    """

    if time_window is None:
        time_window = [get_min_spike_time(units), get_max_spike_time(units)]

    if units_window is None:
        units_window = [0, len(units)]

    if order is None:
        order = np.arange(len(units), dtype='int')

    if progress_bar:
        this_iter = ProgressBar(order, desc='reading spike data', leave=False)
        progress_bar = this_iter.container
    else:
        this_iter = order
    data = []
    for unit in this_iter:
        data.append(get_spike_times(units, unit, time_window))

    if show_obs_intervals:
        unobserved_intervals_list = get_unobserved_intervals(units, time_window, order)
    else:
        unobserved_intervals_list = None

    ax = plot_grouped_events(data, time_window, group_inds=group_inds, labels=labels, show_legend=show_legend,
                             offset=units_window[0], unobserved_intervals_list=unobserved_intervals_list,
                             progress_bar=progress_bar)
    ax.set_ylabel('unit #')

    return ax


class RasterWidget(widgets.HBox):
    def __init__(self, units: Units, time_window_controller=None, group_by=None):
        super().__init__()

        self.units = units

        if time_window_controller is None:
            self.tmin = get_min_spike_time(units)
            self.tmax = get_max_spike_time(units)
            self.time_window_controller = StartAndDurationController(tmin=self.tmin, tmax=self.tmax, start=self.tmin,
                                                                     duration=5)
        else:
            self.time_window_controller = time_window_controller
            self.tmin = self.time_window_controller.vmin
            self.tmax = self.time_window_controller.vmax

        self.gas = self.make_group_and_sort(group_by=group_by)

        self.progress_bar = widgets.HBox()

        self.controls = dict(
            units=fixed(self.units),
            time_window=self.time_window_controller,
            gas=self.gas,
            progress_bar=fixed(self.progress_bar)
        )

        out_fig = interactive_output(show_session_raster, self.controls)

        self.children = [
                self.gas,
                widgets.VBox(
                    children=[
                        self.time_window_controller,
                        self.progress_bar,
                        out_fig,
                    ],
                    layout=Layout(width="100%")
                )
            ]

        self.layout = Layout(width="100%")

    def make_group_and_sort(self, group_by=None):
        return GroupAndSortController(self.units, group_by=group_by)


def show_decomposition_series(node, **kwargs):
    # Use Rendering... as a placeholder
    ntabs = 2
    children = [widgets.HTML('Rendering...') for _ in range(ntabs)]

    def on_selected_index(change):
        # Click on Traces Tab
        if change.new == 1 and isinstance(change.owner.children[1], widgets.HTML):
            widget_box = show_decomposition_traces(node)
            children[1] = widget_box
            change.owner.children = children

    field_lay = widgets.Layout(max_height='40px', max_width='500px',
                               min_height='30px', min_width='130px')
    vbox = []
    for key, val in node.fields.items():
        lbl_key = widgets.Label(key+':', layout=field_lay)
        lbl_val = widgets.Label(str(val), layout=field_lay)
        vbox.append(widgets.HBox(children=[lbl_key, lbl_val]))
    children[0] = widgets.VBox(vbox)

    tab_nest = widgets.Tab()
    tab_nest.children = children
    tab_nest.set_title(0, 'Fields')
    tab_nest.set_title(1, 'Traces')
    tab_nest.observe(on_selected_index, names='selected_index')
    return tab_nest


def show_decomposition_traces(node: DecompositionSeries):
    # Produce figure
    def control_plot(x0, x1, ch0, ch1):
        fig, ax = plt.subplots(nrows=nBands, ncols=1, sharex=True, figsize=(14, 7))
        for bd in range(nBands):
            data = node.data[x0:x1, ch0:ch1+1, bd]
            xx = np.arange(x0, x1)
            mu_array = np.mean(data, 0)
            sd_array = np.std(data, 0)
            offset = np.mean(sd_array) * 5
            yticks = [i*offset for i in range(ch1+1-ch0)]
            for i in range(ch1+1-ch0):
                ax[bd].plot(xx, data[:, i] - mu_array[i] + yticks[i])
            ax[bd].set_ylabel('Ch #', fontsize=20)
            ax[bd].set_yticks(yticks)
            ax[bd].set_yticklabels([str(i) for i in range(ch0, ch1+1)])
            ax[bd].tick_params(axis='both', which='major', labelsize=16)
        ax[bd].set_xlabel('Time [ms]', fontsize=20)
        return fig

    nSamples = node.data.shape[0]
    nChannels = node.data.shape[1]
    nBands = node.data.shape[2]
    fs = node.rate

    # Controls
    field_lay = widgets.Layout(max_height='40px', max_width='100px',
                               min_height='30px', min_width='70px')
    x0 = widgets.BoundedIntText(value=0, min=0, max=int(1000*nSamples/fs-100),
                                layout=field_lay)
    x1 = widgets.BoundedIntText(value=nSamples, min=100, max=int(1000*nSamples/fs),
                                layout=field_lay)
    ch0 = widgets.BoundedIntText(value=0, min=0, max=int(nChannels-1), layout=field_lay)
    ch1 = widgets.BoundedIntText(value=10, min=0, max=int(nChannels-1), layout=field_lay)

    controls = {
        'x0': x0,
        'x1': x1,
        'ch0': ch0,
        'ch1': ch1
    }
    out_fig = widgets.interactive_output(control_plot, controls)

    # Assemble layout box
    lbl_x = widgets.Label('Time [ms]:', layout=field_lay)
    lbl_ch = widgets.Label('Ch #:', layout=field_lay)
    lbl_blank = widgets.Label('    ', layout=field_lay)
    hbox0 = widgets.HBox(children=[lbl_x, x0, x1, lbl_blank, lbl_ch, ch0, ch1])
    vbox = widgets.VBox(children=[hbox0, out_fig])
    return vbox


class PSTHWidget(widgets.VBox):
    def __init__(self, units: Units, unit_index=0, unit_controller=None, sigma_in_secs=.05, ntt=1000):

        self.units = units

        super().__init__()

        self.trials = self.get_trials()
        if self.trials is None:
            self.children = [widgets.HTML('No trials present')]
            return

        if unit_controller is None:
            nunits = len(units['spike_times'].data)
            unit_controller = widgets.Dropdown(options=[x for x in range(nunits)], value=unit_index, description='unit')

        trial_event_controller = make_trial_event_controller(self.trials)
        before_slider = widgets.FloatSlider(.5, min=0, max=5., description='before (s)', continuous_update=False)
        after_slider = widgets.FloatSlider(2., min=0, max=5., description='after (s)', continuous_update=False)

        self.gas = self.make_group_and_sort(window=False)

        self.controls = dict(
            units=fixed(units),
            trials=fixed(self.trials),
            sigma_in_secs=fixed(sigma_in_secs),
            ntt=fixed(ntt),
            index=unit_controller,
            after=after_slider,
            before=before_slider,
            start_label=trial_event_controller,
            gas=self.gas,
            #progress_bar=fixed(progress_bar)
        )

        out_fig = interactive_output(trials_psth, self.controls)

        self.children = [
            widgets.HBox([
                self.gas,
                widgets.VBox([
                    unit_controller,
                    trial_event_controller,
                    before_slider,
                    after_slider,
                ])
            ]),
            out_fig
        ]

    def get_trials(self):
        return self.units.get_ancestor('NWBFile').trials

    def make_group_and_sort(self, window=None):
        return GroupAndSortController(self.trials, window=window)


def trials_psth(units: pynwb.misc.Units, index, start_label='start_time',
                before=0., after=1., order=None, group_inds=None, labels=None,
                sigma_in_secs=0.05, ntt=1000, progress_bar=None, trials=None):
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    index: int
        Index of unit
    start_label: str, optional
        Trial column name to align on
    before: float
        Time before that event (should be positive)
    after: float
        Time after that event
    order
    group_inds
    labels
    sigma_in_secs: float, optional
        standard deviation of gaussian kernel
    ntt:
        Number of time points to use for smooth curve

    Returns
    -------
    matplotlib.Figure

    """
    if trials is None:
        trials = units.get_ancestor('NWBFile').trials

    data = align_by_time_intervals(units, index, trials, start_label, start_label, before, after, order,
                                   progress_bar=progress_bar)
    # expanded data so that gaussian smoother uses larger window than is viewed
    expanded_data = align_by_time_intervals(units, index, trials, start_label, start_label,
                                            before + sigma_in_secs * 4,
                                            after + sigma_in_secs * 4,
                                            order,
                                            progress_bar=progress_bar)

    fig, axs = plt.subplots(2, 1, figsize=(10, 10))

    show_psth_raster(data, before, after, group_inds, labels, ax=axs[0], progress_bar=progress_bar)

    axs[0].set_title('PSTH for unit {}'.format(index))
    axs[0].set_xticks([])
    axs[0].set_xlabel('')

    show_psth_smoothed(expanded_data, axs[1], before, after, group_inds,
                       sigma_in_secs=sigma_in_secs, ntt=ntt)
    return fig


def show_psth_smoothed(data, ax, before, after, group_inds=None, sigma_in_secs=.05, ntt=1000,
                       align_line_color=(.7, .7, .7)):

    all_data = np.hstack(data)
    if not len(all_data):
        return
    tt = np.linspace(min(all_data), max(all_data), ntt)
    smoothed = np.array([compute_smoothed_firing_rate(x, tt, sigma_in_secs) for x in data])

    if group_inds is None:
        group_inds = np.zeros((len(smoothed)))
    group_stats = []
    for group in range(len(np.unique(group_inds))):
        this_mean = np.mean(smoothed[group_inds == group], axis=0)
        err = scipy.stats.sem(smoothed[group_inds == group], axis=0)
        group_stats.append({'mean': this_mean,
                            'lower': this_mean - 2 * err,
                            'upper': this_mean + 2 * err})
    for stats, color in zip(group_stats, color_wheel):
        ax.plot(tt, stats['mean'], color=color)
        ax.fill_between(tt, stats['lower'], stats['upper'], alpha=.2, color=color)
    ax.set_xlim([-before, after])
    ax.set_ylabel('firing rate (Hz)')
    ax.set_xlabel('time (s)')

    ax.axvline(color=align_line_color)


def plot_grouped_events(data, window, group_inds=None, colors=color_wheel, ax=None, labels=None,
                        show_legend=True, offset=0, unobserved_intervals_list=None, progress_bar=None):
    """

    Parameters
    ----------
    data: array-like
    window: array-like [float, float]
        Time in seconds
    group_inds: array-like dtype=int, optional
    colors: array-like, optional
    ax: plt.Axes, optional
    labels: array-like dtype=str, optional
    show_legend: bool, optional
    offset: number, optional
    unobserved_intervals_list: array-like, optional
    progress_bar: FloatProgress, optional

    Returns
    -------

    """

    data = np.asarray(data)
    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 6))
    if group_inds is not None:
        ugroup_inds = np.unique(group_inds)
        handles = []

        if progress_bar is not None:
            this_iter = ProgressBar(enumerate(ugroup_inds), desc='plotting spikes', leave=False, total=len(ugroup_inds))
            progress_bar = this_iter.container
        else:
            this_iter = enumerate(ugroup_inds)

        for i, ui in this_iter:
            color = colors[ugroup_inds[i] % len(colors)]
            lineoffsets = np.where(group_inds == ui)[0] + offset
            event_collection = ax.eventplot(data[group_inds == ui],
                                            orientation='horizontal',
                                            lineoffsets=lineoffsets,
                                            color=color)
            handles.append(event_collection[0])
        if show_legend:
            ax.legend(handles=handles[::-1], labels=list(labels[ugroup_inds][::-1]), loc='upper left',
                      bbox_to_anchor=(1.01, 1))
    else:
        ax.eventplot(data, orientation='horizontal', color='k', lineoffsets=np.arange(len(data)) + offset)

    if unobserved_intervals_list is not None:
        plot_unobserved_intervals(unobserved_intervals_list, ax, offset=offset)

    ax.set_xlim(window)
    ax.set_xlabel('time (s)')
    ax.set_ylim(np.array([-.5, len(data) - .5]) + offset)
    if len(data) <= 30:
        ax.set_yticks(range(offset, len(data) + offset))

    return ax


def plot_unobserved_intervals(unobserved_intervals_list, ax, offset=0, color=(0.85, 0.85, 0.85)):
    for irow, unobs_intervals in enumerate(unobserved_intervals_list):
        rects = [Rectangle((i_interval[0], irow - .5 + offset),
                           i_interval[1] - i_interval[0], 1)
                 for i_interval in unobs_intervals]
        pc = PatchCollection(rects, color=color)
        ax.add_collection(pc)


def show_psth_raster(data, before=0.5, after=2.0, group_inds=None, labels=None, ax=None, show_legend=True,
                     align_line_color=(0.7, 0.7, 0.7), progress_bar: FloatProgress = None) -> plt.Axes:
    """

    Parameters
    ----------
    data: array-like
    before: float, optional
    after: float, optional
    group_inds: array-like, optional
    labels: array-like, optional
    ax: plt.Axes, optional
    show_legend: bool, optional
    align_line_color: array-like, optional
        [R, G, B] (0-1)
        Default = [0.7, 0.7, 0.7]
    progress_bar: FloatProgress, optional

    Returns
    -------
    plt.Axes

    """
    if not len(data):
        return ax
    ax = plot_grouped_events(data, [-before, after], group_inds, color_wheel, ax, labels,
                             show_legend=show_legend, progress_bar=progress_bar)
    ax.set_ylabel('trials')
    ax.axvline(color=align_line_color)
    return ax


def raster_grid(units: pynwb.misc.Units, time_intervals: pynwb.epoch.TimeIntervals, index, before, after,
                rows_label=None, cols_label=None, trials_select=None, align_by='start_time') -> plt.Figure:
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    time_intervals: pynwb.epoch.TimeIntervals
    index: int
    before: float
    after: float
    rows_label: str, optional
    cols_label: str, optional
    trials_select: np.array(dtype=bool), optional
    align_by: str, optional

    Returns
    -------
    plt.Figure

    """
    if time_intervals is None:
        raise ValueError('trials must exist (trials cannot be None)')

    if trials_select is None:
        trials_select = np.ones((len(time_intervals),)).astype('bool')

    if rows_label is not None:
        row_vals = time_intervals[rows_label][:]
        urow_vals = np.unique(row_vals[trials_select])
        if urow_vals.dtype == np.float64:
            urow_vals = urow_vals[~np.isnan(urow_vals)]

    else:
        urow_vals = [None]
    nrows = len(urow_vals)

    if cols_label is not None:
        col_vals = time_intervals[cols_label][:]
        ucol_vals = np.unique(col_vals[trials_select])
        if ucol_vals.dtype == np.float64:
            ucol_vals = ucol_vals[~np.isnan(ucol_vals)]

    else:
        ucol_vals = [None]
    ncols = len(ucol_vals)

    fig, axs = plt.subplots(nrows, ncols, sharex=True, sharey=True, squeeze=False, figsize=(10, 10))
    big_ax = create_big_ax(fig)
    for i, row in enumerate(urow_vals):
        for j, col in enumerate(ucol_vals):
            ax = axs[i, j]
            ax_trials_select = trials_select.copy()
            if row is not None:
                ax_trials_select &= row_vals == row
            if col is not None:
                ax_trials_select &= col_vals == col
            ax_trials_select = np.where(ax_trials_select)[0]
            if len(ax_trials_select):
                data = align_by_time_intervals(units, index, time_intervals, align_by, align_by,
                                               before, after, ax_trials_select)
                show_psth_raster(data, before, after, ax=ax)
                ax.set_xlabel('')
                ax.set_ylabel('')
                if ax.is_first_col():
                    ax.set_ylabel(row)
                if ax.is_last_row():
                    ax.set_xlabel(col)

    big_ax.set_xlabel(cols_label, labelpad=50)
    big_ax.set_ylabel(rows_label, labelpad=60)

    return fig


class RasterGridWidget(widgets.VBox):

    def __init__(self, units: Units, unit_index=0):
        super().__init__()

        self.units = units

        self.trials = self.get_trials()
        if self.trials is None:
            self.children = [widgets.HTML('No trials present')]
            return

        groups = list(self.trials.colnames)

        rows_controller = widgets.Dropdown(options=[None] + list(groups), description='rows')
        cols_controller = widgets.Dropdown(options=[None] + list(groups), description='cols')

        trial_event_controller = make_trial_event_controller(self.trials)
        unit_controller = widgets.Dropdown(options=range(len(units['spike_times'].data)), value=unit_index,
                                           description='unit')

        before_slider = widgets.FloatSlider(.1, min=0, max=5., description='before (s)', continuous_update=False)
        after_slider = widgets.FloatSlider(1., min=0, max=5., description='after (s)', continuous_update=False)

        self.controls = {
            'units': fixed(units),
            'time_intervals': fixed(self.trials),
            'index': unit_controller,
            'after': after_slider,
            'before': before_slider,
            'align_by': trial_event_controller,
            'rows_label': rows_controller,
            'cols_label': cols_controller
        }

        self.children = [
            unit_controller,
            rows_controller,
            cols_controller,
            trial_event_controller,
            before_slider,
            after_slider,
        ]

        self.select_trials()

        out_fig = interactive_output(raster_grid, self.controls, self.process_controls)

        self.children = list(self.children) + [out_fig]

    def get_groups(self):
        return infer_categorical_columns(self.trials)

    @staticmethod
    def get_group_vals(dynamic_table, group_by, units_select=()):
        if group_by is None:
            return None
        elif group_by in dynamic_table:
            return dynamic_table[group_by][:][units_select]
        else:
            raise ValueError('{} not found in trials'.format(group_by))

    def get_trials(self):
        return self.units.get_ancestor('NWBFile').trials

    def select_trials(self):
        return

    def process_controls(self, control_states):
        return control_states




import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
from pynwb.misc import DecompositionSeries, AnnotationSeries
from ipywidgets import widgets
from nwbwidgets.misc import show_psth_raster, PSTHWidget, show_decomposition_traces, show_decomposition_series, RasterWidget, \
    show_session_raster, show_annotations, RasterGridWidget, raster_grid
import unittest


def test_show_psth():
    data = np.random.random([6, 50])
    assert isinstance(show_psth_raster(data=data, before=0, after=1), plt.Subplot)


def test_show_annotations():
    timestamps = np.array([0., 1., 2., 3., 4., 5., 6.])
    annotations = AnnotationSeries(name='test_annotations',timestamps=timestamps)
    show_annotations(annotations)

    
class ShowPSTHTestCase(unittest.TestCase):

    def setUp(self):
        """
        Trials must exist.
        """
        start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
        create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

        self.nwbfile = NWBFile(session_description='NWBFile for PSTH',
                               identifier='NWB123',
                               session_start_time=start_time,
                               file_create_date=create_date)

        self.nwbfile.add_unit_column('location', 'the anatomical location of this unit')
        self.nwbfile.add_unit_column('quality', 'the quality for the inference of this unit')

        self.nwbfile.add_unit(spike_times=[2.2, 3.0, 4.5],
                              obs_intervals=[[1, 10]], location='CA1', quality=0.95)
        self.nwbfile.add_unit(spike_times=[2.2, 3.0, 25.0, 26.0],
                              obs_intervals=[[1, 10], [20, 30]], location='CA3', quality=0.85)
        self.nwbfile.add_unit(spike_times=[1.2, 2.3, 3.3, 4.5],
                              obs_intervals=[[1, 10], [20, 30]], location='CA1', quality=0.90)

        self.nwbfile.add_trial_column(name='stim', description='the visual stimuli during the trial')

        self.nwbfile.add_trial(start_time=0.0, stop_time=2.0, stim='person')
        self.nwbfile.add_trial(start_time=3.0, stop_time=5.0, stim='ocean')
        self.nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim='desert')

    def test_psth_widget(self):
        assert isinstance(PSTHWidget(self.nwbfile.units), widgets.Widget)

    def test_raster_widget(self):
        assert isinstance(RasterWidget(self.nwbfile.units), widgets.Widget)

    def test_show_session_raster(self):
        assert isinstance(show_session_raster(self.nwbfile.units), plt.Axes)

    def test_raster_grid_widget(self):
        assert isinstance(RasterGridWidget(self.nwbfile.units), widgets.Widget)

    def test_raster_grid(self):
        trials = self.nwbfile.units.get_ancestor('NWBFile').trials
        assert isinstance(raster_grid(self.nwbfile.units, time_intervals=trials, index=0, before=0.5, after=20.0), plt.Figure)

    
class ShowDecompositionTestCase(unittest.TestCase):

    def setUp(self):
    
        data = np.random.rand(160, 2, 3)

        self.ds = DecompositionSeries(name='Test Decomposition', data=data,
                                      metric='amplitude', rate=1.0)

    def test_show_decomposition_traces(self):

        assert isinstance(show_decomposition_traces(self.ds), widgets.Widget)

    def test_show_decomposition_series(self):

        assert isinstance(show_decomposition_series(self.ds), widgets.Widget)



import numpy as np
from pynwb import TimeSeries
import ipywidgets as widgets
from nwbwidgets.view import default_neurodata_vis_spec
from nwbwidgets.ecephys import show_lfp, show_spectrogram, show_spike_event_series
from pynwb.ecephys import LFP, ElectricalSeries, SpikeEventSeries
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
import unittest


class ShowActivityTestCase(unittest.TestCase):
    
    def setUp(self):
        
        nwbfile = NWBFile('my first synthetic recording', 'EXAMPLE_ID', datetime.now(tzlocal()),
                  experimenter='Dr. Matthew Douglass',
                  lab='Vision Neuroscience Laboratory',
                  institution='University of Middle Earth at the Shire',
                  experiment_description='We recorded from two macaque monkeys during memory-guided saccade task',
                  session_id='LONELYMTL')
        
        device = nwbfile.create_device(name='trodes_rig123')
        
        electrode_group = nwbfile.create_electrode_group(name='tetrode1',
                                                     description='an example tetrode',
                                                     location='somewhere in the hippocampus',
                                                     device=device)
        
        for idx in [1, 2, 3, 4]:
            nwbfile.add_electrode(id=idx,
                                  x=1.0, y=2.0, z=3.0,
                                  imp=float(-idx),
                                  location='CA1', filtering='none',
                                  group=electrode_group)
        
        electrode_table_region = nwbfile.create_electrode_table_region([0, 2], 'the first and third electrodes')
        
        self.electrodes = electrode_table_region


    def test_show_lfp(self):

        rate = 10.0
        np.random.seed(1234)
        data_len = 1000
        ephys_data = np.random.rand(data_len * 2).reshape((data_len, 2))
        ephys_timestamps = np.arange(data_len) / rate

        ephys_ts = ElectricalSeries('test_ephys_data',
                                    ephys_data,
                                    self.electrodes,
                                    timestamps=ephys_timestamps,
                                    resolution=0.001,
                                    comments="This data was randomly generated with numpy, using 1234 as the seed",
                                    description="Random numbers generated with numpy.random.rand")
        

        lfp = LFP(electrical_series=ephys_ts, name='LFP data')

        show_lfp(lfp, default_neurodata_vis_spec)


    def test_show_spike_event_series(self):

        rate = 10.0
        np.random.seed(1234)
        data_len = 1000
        ephys_data = np.random.rand(data_len * 6).reshape((data_len, 2, 3))
        ephys_timestamps = np.arange(data_len) / rate
        ses = SpikeEventSeries(name='test_ephys_data', 
                               data=ephys_data, 
                               timestamps=ephys_timestamps, 
                               electrodes=self.electrodes, 
                               resolution=0.001, 
                               comments="This data was randomly generated with numpy, using 1234 as the seed", 
                               description="Random numbers generated with numpy.random.rand")

        assert isinstance(show_spike_event_series(ses),widgets.Widget)

     
    
def test_show_spectrogram():
    
    data = np.random.rand(160,12)
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0.0, rate=1.0)
    
    channel = 3
    show_spectrogram(ts,channel=channel)
    



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from pynwb import TimeSeries
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
from ipywidgets import widgets
from pynwb.core import DynamicTable
from pynwb.file import Subject
from nwbwidgets.view import default_neurodata_vis_spec
from pynwb import ProcessingModule
from pynwb.behavior import Position, SpatialSeries
from nwbwidgets.base import show_neurodata_base,processing_module, nwb2widget, show_text_fields, \
fig2widget, vis2widget, show_fields, show_dynamic_table, df2accordion, lazy_show_over_data
import unittest
import pytest
    
    
def test_show_neurodata_base():
    start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
    create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())
    
    nwbfile = NWBFile(session_description='demonstrate NWBFile basics',  
                      identifier='NWB123',  
                      session_start_time=start_time,  
                      file_create_date=create_date,
                      related_publications='https://doi.org/10.1088/1741-2552/aaa904',
                      experimenter='Dr. Pack')
    
    assert isinstance(show_neurodata_base(nwbfile,default_neurodata_vis_spec), widgets.Widget)
    

def test_show_text_fields():
    data = np.random.rand(160,3)
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0.0, rate=1.0)
    assert isinstance(show_text_fields(ts), widgets.Widget)
    
    
class ProcessingModuleTestCase(unittest.TestCase):
    
    def setUp(self):
        spatial_series = SpatialSeries(name='position',
                                   data=np.linspace(0, 1, 20),
                                   rate=50.,
                                   reference_frame='starting gate')
        self.position = Position(spatial_series=spatial_series)
    
    def test_processing_module(self):

        start_time = datetime(2020, 1, 29, 11, tzinfo=tzlocal())
        nwbfile = NWBFile(session_description='Test Session',  
                      identifier='NWBPM',  
                      session_start_time=start_time)

        behavior_module = ProcessingModule(name='behavior',
                                                       description='preprocessed behavioral data')
        nwbfile.add_processing_module(behavior_module)

        nwbfile.processing['behavior'].add(self.position)

        processing_module(nwbfile.processing['behavior'], default_neurodata_vis_spec)

    def test_nwb2widget(self):

        nwb2widget(self.position, default_neurodata_vis_spec)


def test_fig2widget():
    
    data = np.random.rand(160, 3)
    fig = plt.figure(figsize=(10, 5))
    plt.plot(data)
    
    assert isinstance(fig2widget(fig), widgets.Widget)


class Test_vis2widget:
    def test_vis2widget_input_widget(self):

        wg = widgets.IntSlider(
            value=7,
            min=0,
            max=10,
            step=1,
            description='Test:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='d')

        assert isinstance(vis2widget(wg), widgets.Widget)

    def test_vis2widget_input_figure(self):

        data = np.random.rand(160,3)

        fig=plt.figure(figsize=(10, 5))
        plt.plot(data)

        assert isinstance(vis2widget(fig), widgets.Widget)
        
    def test_vis2widget_input_other(self):
        data = np.random.rand(160,3)
        with pytest.raises(ValueError, match="unsupported vis type"):
            vis2widget(data)


def test_show_subject():
    node = Subject(age='8', sex='m', species='macaque')
    show_fields(node)


def test_show_dynamic_table():
    d = {'col1': [1, 2], 'col2': [3, 4]}
    DT = DynamicTable.from_dataframe(df=pd.DataFrame(data=d), 
                                     name='Test Dtable', 
                                     table_description='no description')
    show_dynamic_table(DT)


def test_df2accordion():
    df = pd.DataFrame(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                     columns=['a', 'b', 'c'])
    def func_fig(data):
        fig=plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig
    df2accordion(df=df,by='a',func=func_fig)
    
    
def test_df2accordion_single():
    df = pd.DataFrame(np.array([1]),
                     columns=['a'])
    def func_fig(data):
        fig=plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig
    df2accordion(df=df,by='a',func=func_fig)
    
    
def test_lazy_show_over_data():
    list_ = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    def func_fig(data):
        fig=plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig
    assert isinstance(lazy_show_over_data(list_=list_,func_=func_fig),widgets.Widget)



import matplotlib.pyplot as plt
from matplotlib import gridspec


def create_big_ax(fig):
    big_ax = plt.Subplot(fig, gridspec.GridSpec(1, 1)[0])
    fig.add_subplot(big_ax)
    [sp.set_visible(False) for sp in big_ax.spines.values()]
    big_ax.set_xticks([])
    big_ax.set_yticks([])
    big_ax.patch.set_facecolor('none')

    return big_ax


def grid_sps(shape, subplot_spec=None, fig=None):
    """
    Create subplot_spec from pandas.DataFrame

    Parameters
    ----------
    shape: tuple
    subplot_spec: GridSpec, optional
    fig: matplotlib.pyplot.Figure, optional

    Returns
    -------

    """

    if fig is None:
        fig = plt.gcf()

    if subplot_spec is not None:
        gs = gridspec.GridSpecFromSubplotSpec(shape[0], shape[1], subplot_spec=subplot_spec)
        big_ax = plt.Subplot(fig, subplot_spec)
    else:
        gs = gridspec.GridSpec(shape[0], shape[1], figure=fig)
        big_ax = plt.Subplot(fig, gridspec.GridSpec(1, 1)[0])
    fig.add_subplot(big_ax)

    [sp.set_visible(False) for sp in big_ax.spines.values()]
    big_ax.set_xticks([])
    big_ax.set_yticks([])
    big_ax.patch.set_facecolor('none')

    return fig, big_ax, gs




import numpy as np
import matplotlib.pyplot as plt
from pynwb.icephys import IntracellularElectrode
from pynwb.base import TimeSeries
from ndx_icephys_meta.icephys import Sweeps,IntracellularRecordings
from nwbwidgets.icephys import show_single_sweep_sequence
from pynwb.device import Device




def test_show_single_sweep_sequence():
    
    device = Device(name='Axon Patch-Clamp')
    electrode = IntracellularElectrode(name='Patch Clamp',device=device, 
                                       description='whole-cell')
    
    stimulus_data = np.random.rand(160,2)
    stimulus = TimeSeries(name='test_timeseries', data=stimulus_data, unit='m', starting_time=0.0, rate=1.0)
    response_data = np.random.rand(160,2)
    response = TimeSeries(name='test_timeseries', data=response_data, unit='m', starting_time=0.0, rate=1.0)
    
    icr = IntracellularRecordings()
    icr.add_recording(electrode=electrode, stimulus_start_index=0, stimulus_index_count=100, 
                      stimulus=stimulus, response_start_index=0, response_index_count=100, 
                      response=response)
    
    sweeps_table = Sweeps(intracellular_recordings_table=icr)
    assert isinstance(show_single_sweep_sequence(sweeps_table),plt.Figure)




from nwbwidgets import view
import numpy as np
import matplotlib.pyplot as plt
from ipywidgets import widgets
from pynwb.behavior import Position, SpatialSeries, BehavioralEvents
from nwbwidgets import base
from .utils.timeseries import get_timeseries_tt, get_timeseries_in_units


def show_position(node: Position, neurodata_vis_spec: dict):

    if len(node.spatial_series.keys()) == 1:
        for value in node.spatial_series.values():
            return view.nwb2widget(value, neurodata_vis_spec=neurodata_vis_spec)
    else:
        return view.nwb2widget(node.spatial_series, neurodata_vis_spec=neurodata_vis_spec)


def show_behavioral_events(beh_events: BehavioralEvents, neurodata_vis_spec: dict):
    return base.dict2accordion(beh_events.time_series, neurodata_vis_spec, ls='', marker='|')


def show_spatial_series_over_time(node: SpatialSeries, **kwargs):

    text_widget = base.show_text_fields(
        node, exclude=('timestamps_unit', 'comments', 'data', 'timestamps', 'interval'))

    data, unit = get_timeseries_in_units(node)

    if len(data.shape) == 1:
        ndims = 1
    else:
        ndims = data.shape[1]

    tt = get_timeseries_tt(node)

    if ndims == 1:
        fig, ax = plt.subplots()
        ax.plot(tt, data, **kwargs)
        ax.set_xlabel('t (sec)')
        if unit:
            ax.set_ylabel('x ({})'.format(unit))
        else:
            ax.set_ylabel('x')

    else:
        fig, axs = plt.subplots(ndims, 1, sharex=True)

        for i, (ax, dim_label) in enumerate(zip(axs, ('x', 'y', 'z'))):
            ax.plot(tt, data[:, i], **kwargs)
            if unit:
                ax.set_ylabel(dim_label + ' ({})'.format(unit))
            else:
                ax.set_ylabel(dim_label)
        ax.set_xlabel('t (sec)')

    return widgets.HBox([text_widget, base.fig2widget(fig)])


def show_spatial_series(node: SpatialSeries, **kwargs):

    data, unit = get_timeseries_in_units(node)
    tt = get_timeseries_tt(node)

    if len(data.shape) == 1:
        fig, ax = plt.subplots()
        ax.plot(tt, data, **kwargs)
        ax.set_xlabel('t (sec)')
        if unit:
            ax.set_xlabel('x ({})'.format(unit))
        else:
            ax.set_xlabel('x')
        ax.set_ylabel('x')

    elif data.shape[1] == 2:
        fig, ax = plt.subplots()
        ax.plot(data[:, 0], data[:, 1], **kwargs)
        if unit:
            ax.set_xlabel('x ({})'.format(unit))
            ax.set_ylabel('y ({})'.format(unit))
        else:
            ax.set_xlabel('x')
            ax.set_ylabel('y')
        ax.axis('equal')

    elif data.shape[1] == 3:
        import ipyvolume.pylab as p3

        fig = p3.figure()
        p3.scatter(data[:, 0], data[:, 1], data[:, 2], **kwargs)
        p3.xlim(np.min(data[:, 0]), np.max(data[:, 0]))
        p3.ylim(np.min(data[:, 1]), np.max(data[:, 1]))
        p3.zlim(np.min(data[:, 2]), np.max(data[:, 2]))

    else:
        raise NotImplementedError

    return fig



from ipywidgets import widgets, Layout, ValueWidget, link, HBox
from ipywidgets.widgets.widget_description import DescriptionWidget
import numpy as np
from hdmf.common import DynamicTable
from .utils.dynamictable import group_and_sort, infer_categorical_columns
from .utils.pynwb import robust_unique
from typing import Iterable

from tqdm.notebook import tqdm as tqdm_notebook


class RangeController(widgets.HBox, ValueWidget, DescriptionWidget):

    def __init__(self, vmin, vmax, start_value=None, dtype='float', description='time window (s)',
                 orientation='horizontal', **kwargs):

        if orientation not in ('horizontal', 'vertical'):
            ValueError('Unrecognized orientation: {}'.format(orientation))

        self.vmin = vmin
        self.vmax = vmax
        self.start_value = start_value
        self.orientation = orientation
        self.dtype = dtype

        super().__init__()

        self.slider = self.make_range_slider(description=description, **kwargs)

        [link((self.slider, attr), (self, attr)) for attr in ('value', 'description')]

        if self.orientation == 'horizontal':
            self.to_start_button = widgets.Button(description='◀◀', layout=Layout(width='55px'))
            self.backwards_button = widgets.Button(description='◀', layout=Layout(width='40px'))
            self.forward_button = widgets.Button(description='▶', layout=Layout(width='40px'))
            self.to_end_button = widgets.Button(description='▶▶', layout=Layout(width='55px'))
        else:  # vertical
            self.to_end_button = widgets.Button(description='▲▲', layout=Layout(width='50px'))
            self.forward_button = widgets.Button(description='▲', layout=Layout(width='50px'))
            self.backwards_button = widgets.Button(description='▼', layout=Layout(width='50px'))
            self.to_start_button = widgets.Button(description='▼▼', layout=Layout(width='50px'))

        self.to_start_button.on_click(self.move_start)
        self.backwards_button.on_click(self.move_down)
        self.forward_button.on_click(self.move_up)
        self.to_end_button.on_click(self.move_end)

        self.children = self.get_children()

    def get_children(self):
        if self.orientation == 'horizontal':
            return [
                self.slider,
                self.to_start_button,
                self.backwards_button,
                self.forward_button,
                self.to_end_button
            ]
        elif self.orientation == 'vertical':
            return [widgets.VBox([
                self.slider,
                self.to_end_button,
                self.forward_button,
                self.backwards_button,
                self.to_start_button,
            ],
                layout=widgets.Layout(display='flex',
                                      flex_flow='column',
                                      align_items='center')
            )]
        else:
            raise ValueError('Unrecognized orientation: {}'.format(self.orientation))

    def make_range_slider(self, **kwargs):
        """

        Parameters
        ----------
        kwargs: passed into RangeSlider constructor

        Returns
        -------

        """

        slider_kwargs = dict(
            value=self.start_value,
            min=self.vmin,
            max=self.vmax,
            continuous_update=False,
            readout=True,
            style={'description_width': 'initial'},
            orientation=self.orientation
        )

        if self.dtype == 'float':
            slider_kwargs.update(
                readout_format='.1f',
                step=0.1,
                description='time window (s)',
                layout=Layout(width='100%')
            )
            slider_kwargs.update(kwargs)
            return widgets.FloatRangeSlider(**slider_kwargs)
        elif self.dtype == 'int':
            slider_kwargs.update(
                description='unit window',
                layout=Layout(height='100%')
            )
            slider_kwargs.update(kwargs)
            return widgets.IntRangeSlider(**slider_kwargs)
        else:
            raise ValueError('Unrecognized dtype: {}'.format(self.dtype))

    def move_up(self, change):
        value_range = self.value[1] - self.value[0]
        if self.value[1] + value_range < self.vmax:
            self.value = (self.value[0] + value_range, self.value[1] + value_range)
        else:
            self.move_end(change)

    def move_down(self, change):
        value_range = self.value[1] - self.value[0]
        if self.value[0] - value_range > self.vmin:
            self.value = (self.value[0] - value_range, self.value[1] - value_range)
        else:
            self.move_start(change)

    def move_start(self, change):
        value_range = self.value[1] - self.value[0]
        self.value = (self.vmin, self.vmin + value_range)

    def move_end(self, change):
        value_range = self.value[1] - self.value[0]
        self.value = (self.vmax - value_range, self.vmax)


class StartAndDurationController(HBox, ValueWidget, DescriptionWidget):
    """
    Can be used in place of the RangeController.
    """
    def __init__(self, tmax, tmin=0, start_value=None, duration=1., dtype='float', description='window (s)',
                 **kwargs):

        self.tmin = tmin
        self.tmax = tmax
        self.start_value = start_value
        self.dtype = dtype

        self.slider = widgets.FloatSlider(
            value=start_value,
            min=tmin,
            max=tmax,
            step=0.01,
            description=description,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='.2f',
            style={'description_width': 'initial'},
            layout=Layout(width='100%'))

        self.duration = widgets.BoundedFloatText(
            value=duration,
            min=0,
            max=tmax - tmin,
            step=0.1,
            description='duration (s):',
            style={'description_width': 'initial'},
            layout=Layout(max_width='140px')
        )

        super().__init__()
        link((self.slider, 'description'), (self, 'description'))

        self.value = (self.slider.value, self.slider.value + self.duration.value)

        self.forward_button = widgets.Button(description='▶', layout=Layout(width='50px'))
        self.forward_button.on_click(self.move_up)

        self.backwards_button = widgets.Button(description='◀', layout=Layout(width='50px'))
        self.backwards_button.on_click(self.move_down)

        self.children = [self.slider, self.duration, self.backwards_button, self.forward_button]

        self.slider.observe(self.monitor_slider)
        self.duration.observe(self.monitor_duration)

    def monitor_slider(self, change):
        if 'new' in change:
            if isinstance(change['new'], dict):
                if 'value' in change['new']:
                    value = change['new']['value']
                else:
                    return
            else:
                value = change['new']
        if value + self.duration.value > self.tmax:
            self.slider.value = self.tmax - self.duration.value
        else:
            self.value = (value, value + self.duration.value)

    def monitor_duration(self, change):
        if 'new' in change:
            if isinstance(change['new'], dict):
                if 'value' in change['new']:
                    value = change['new']['value']
                    if self.slider.value + value > self.tmax:
                        self.slider.value = self.tmax - value
                    self.value = (self.slider.value, self.slider.value + value)

    def move_up(self, change):
        if self.slider.value + 2 * self.duration.value < self.tmax:
            self.slider.value += self.duration.value
        else:
            self.slider.value = self.tmax - self.duration.value

    def move_down(self, change):
        if self.slider.value - self.duration.value > self.tmin:
            self.slider.value -= self.duration.value
        else:
            self.slider.value = self.tmin


class AbstractGroupAndSortController(widgets.VBox, ValueWidget):
    """
    Defines the abstract type for GroupAndSortController objects. These classes take in a DynamicTable objects
    and broadcast a `value` of the form
    dict(
        order=array-like(uint),
        group_inds=array-like(uint) | None,
        labels=array-like(str) | None
    )
    """
    def __init__(self, dynamic_table: DynamicTable):
        super().__init__()

        self.dynamic_table = dynamic_table
        self.nitems = len(self.dynamic_table.id)

        self.group_vals = None
        self.group_by = None
        self.group_select = None
        self.limit = None
        self.desc = False
        self.order_by = None
        self.order_vals = None
        self.window = None


class GroupAndSortController(AbstractGroupAndSortController):
    def __init__(self, dynamic_table: DynamicTable, group_by=None, window=None, start_discard_rows=None):
        """

        Parameters
        ----------
        dynamic_table
        group_by
        window: None or bool,
        """
        super().__init__(dynamic_table)

        groups = self.get_groups()

        self.discard_rows = start_discard_rows

        self.limit_bit = widgets.BoundedIntText(value=50, min=0, max=99999, disabled=True,
                                                layout=Layout(max_width='70px'))
        self.limit_bit.observe(self.limit_bit_observer)

        self.limit_cb = widgets.Checkbox(description='limit', style={'description_width': 'initial'}, disabled=True,
                                         indent=False, layout=Layout(max_width='70px'))
        self.limit_cb.observe(self.limit_cb_observer)

        self.order_dd = widgets.Dropdown(options=[None] + list(groups), description='order by',
                                         layout=Layout(max_width='120px'), style={'description_width': 'initial'})
        self.order_dd.observe(self.order_dd_observer)

        self.ascending_dd = widgets.Dropdown(options=['ASC', 'DESC'], disabled=True,
                                             layout=Layout(max_width='70px'))
        self.ascending_dd.observe(self.ascending_dd_observer)

        range_controller_max = min(30, self.nitems)
        if window is None:
            self.range_controller = RangeController(0, self.nitems, start_value=(0, range_controller_max), dtype='int', description='units',
                                                    orientation='vertical')
            self.range_controller.observe(self.range_controller_observer)
            self.window = self.range_controller.value
        elif window is False:
            self.window = (0, self.nitems)
            self.range_controller = widgets.HTML('')

        self.group_sm = widgets.SelectMultiple(layout=Layout(max_width='100px'), disabled=True, rows=1)
        self.group_sm.observe(self.group_sm_observer)

        if group_by is None:
            self.group_dd = widgets.Dropdown(options=[None] + list(groups), description='group by',
                                             style={'description_width': 'initial'}, layout=Layout(width='90%'))
            self.group_dd.observe(self.group_dd_observer)
        else:
            self.group_dd = None
            self.set_group_by(group_by)

        self.children = self.get_children()
        self.layout = Layout(width='280px')
        self.update_value()

    def get_children(self):
        children = [
            widgets.HBox(children=(self.group_sm, self.range_controller)),
            widgets.HBox(children=(self.limit_cb, self.limit_bit), layout=Layout(max_width='90%')),
            widgets.HBox(children=(self.order_dd, self.ascending_dd), layout=Layout(max_width='90%')),
        ]

        if self.group_dd:
            children.insert(0, self.group_dd)

        return children

    def set_group_by(self, group_by):
        self.group_by = group_by
        self.group_vals = self.get_group_vals(by=group_by)
        group_vals = self.group_vals
        if self.discard_rows is not None:
            group_vals = group_vals[~np.isin(np.arange(len(group_vals), dtype='int'), self.discard_rows)]
        if self.group_vals.dtype == np.float:
            group_vals = group_vals[~np.isnan(group_vals)]
        groups = np.unique(group_vals)
        self.group_sm.options = tuple(groups[::-1])
        self.group_sm.value = self.group_sm.options
        self.group_sm.disabled = False
        self.group_sm.rows = min(len(groups), 20)
        self.limit_cb.disabled = False
        self.group_and_sort()

    def group_dd_observer(self, change):
        """group dropdown observer"""
        if change['name'] == 'value':
            group_by = change['new']
            if group_by in ('None', '', None):
                self.limit_bit.disabled = True
                self.limit_cb.disabled = True
                self.group_vals = None
                self.group_by = None
                self.limit = None
                self.limit_cb.value = False

                if hasattr(self.range_controller, 'slider'):
                    self.range_controller.slider.max = len(self.dynamic_table) - len(self.discard_rows)
            else:
                self.set_group_by(group_by)

            self.update_value()

    def limit_bit_observer(self, change):
        """limit bounded int text observer"""
        if change['name'] == 'value':
            limit = self.limit_bit.value
            self.limit = limit
            self.update_value()

    def limit_cb_observer(self, change):
        """limit checkbox observer"""
        if change['name'] == 'value':
            if self.limit_cb.value and self.group_by is not None:
                self.limit_bit.disabled = False
                self.limit = self.limit_bit.value
            else:
                self.limit_bit.disabled = True
                self.limit = None
            self.update_value()

    def order_dd_observer(self, change):
        """order dropdown observer"""
        if change['name'] == 'value':
            self.order_by = self.order_dd.value

            order_vals = self.get_group_vals(by=self.order_by)
            # convert to ints. This is mainly for handling strings
            _, order_vals = np.unique(order_vals, return_inverse=True)

            if self.desc:  # if descend is on, invert order.
                order_vals *= -1

            self.order_vals = order_vals

            self.ascending_dd.disabled = self.order_dd.value is None
            self.update_value()

    def ascending_dd_observer(self, change):
        """ascending dropdown observer"""
        if change['name'] == 'value':
            if change['new'] == 'ASC':
                self.desc = False
                self.order_vals *= -1
            else:
                self.desc = True
                self.order_vals *= -1
            self.update_value()

    def group_sm_observer(self, change):
        """group SelectMultiple observer"""
        if change['name'] == 'value' and not self.group_sm.disabled:
            self.group_select = change['new']
            value_before = self.window
            self.group_and_sort()
            if hasattr(self.range_controller, 'slider') and not self.range_controller.slider.value == value_before:
                pass  # do nothing, value was updated automatically
            else:
                self.update_value()

    def range_controller_observer(self, change):
        self.window = self.range_controller.value
        self.update_value()

    def get_groups(self):
        return infer_categorical_columns(self.dynamic_table)

    def get_group_vals(self, by, units_select=()):
        """Get the values of the group_by variable

        Parameters
        ----------
        by
        units_select

        Returns
        -------

        """
        if by is None:
            return None
        elif by in self.dynamic_table:
            return self.dynamic_table[by][:][units_select]
        else:
            raise ValueError('column {} not in DynamicTable {}'.format(by, self.dynamic_table))

    def get_orderable_cols(self):
        candidate_cols = [x for x in self.units.colnames
                          if not isinstance(self.units[x][0], Iterable) or
                          isinstance(self.units[x][0], str)]
        return [x for x in candidate_cols if len(robust_unique(self.units[x][:])) > 1]

    def group_and_sort(self):
        if self.group_vals is None and self.order_vals is None:
            self.order_vals = np.arange(self.nitems).astype('int')

        order, group_inds, labels = group_and_sort(
            group_vals=self.group_vals,
            group_select=self.group_select,
            discard_rows=self.discard_rows,
            order_vals=self.order_vals,
            limit=self.limit
        )

        if hasattr(self.range_controller, 'slider'):
            self.range_controller.slider.max = len(order)

        # apply window
        if self.window is not None:
            order = order[self.window[0]:self.window[1]]
            if group_inds is not None:
                group_inds = group_inds[self.window[0]:self.window[1]]

        return order, group_inds, labels

    def update_value(self):

        order, group_inds, labels = self.group_and_sort()
        self.value = dict(order=order, group_inds=group_inds, labels=labels)


class ProgressBar(tqdm_notebook):

    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        self.container.children[0].layout = Layout(width='80%')


def make_trial_event_controller(trials):
    trial_events = ['start_time']
    if not np.all(np.isnan(trials['stop_time'].data)):
        trial_events.append('stop_time')
    trial_events += [x.name for x in trials.columns if
                     (('_time' in x.name) and (x.name not in ('start_time', 'stop_time')))]
    trial_event_controller = widgets.Dropdown(options=trial_events,
                                              value='start_time',
                                              description='align to: ')
    return trial_event_controller



import numpy as np
import matplotlib.pyplot as plt
from pynwb.base import TimeSeries
import ipywidgets as widgets
from pynwb.image import RGBImage,GrayscaleImage,IndexSeries,ImageSeries
from nwbwidgets.image import show_rbg_image,show_grayscale_image,show_index_series,show_image_series
from nwbwidgets.view import default_neurodata_vis_spec


def test_show_rbg_image():
    
    data = np.random.rand(2700).reshape((30,30,3))
    rgb_image = RGBImage(name='test_image',data=data)
    
    assert isinstance(show_rbg_image(rgb_image),plt.Figure)



def test_show_grayscale_image():
    
    data = np.random.rand(900).reshape((30,30))
    grayscale_image = GrayscaleImage(name='test_image',data=data)
    
    assert isinstance(show_grayscale_image(grayscale_image),plt.Figure)

    
def test_show_index_series():
    
    data = np.array([12,14,16,18,20,22,24,26])
    indexed_timeseries = TimeSeries(name='Index Series time data',
                                data=np.random.rand(800).reshape((8,10,10)),rate=1.)
    index_series = IndexSeries(name='Sample Index Series', data=data, 
                               indexed_timeseries=indexed_timeseries,rate=1.)

    assert isinstance(show_index_series(index_series,default_neurodata_vis_spec),widgets.Widget)
    
    
def test_show_image_series():
    
    data=np.random.rand(800).reshape((8,10,10))
    image_series = ImageSeries(name='Image Series',data=data,rate=1.)
    
    assert isinstance(show_image_series(image_series,default_neurodata_vis_spec),widgets.Widget)



from ipywidgets import Output
from ipywidgets.widgets.interaction import show_inline_matplotlib_plots, clear_output
import asyncio


def unpack_controls(controls, process_controls=lambda x: x):
    control_states = {}
    for k, v in controls.items():
        # if the value is a dict, add those individually
        if isinstance(v.value, dict):
            control_states.update(v.value)
        else:
            control_states[k] = v.value
    kwargs = process_controls(control_states)
    return kwargs


def interactive_output(f, controls, process_controls=lambda x: x):
    """Connect widget controls to a function.

    This function does not generate a user interface for the widgets (unlike `interact`).
    This enables customisation of the widget user interface layout.
    The user interface layout must be defined and displayed manually.
    """

    out = Output()

    def observer(change):
        show_inline_matplotlib_plots()
        with out:
            clear_output(wait=True)
            f(**unpack_controls(controls, process_controls))
            show_inline_matplotlib_plots()
    for k, w in controls.items():
        w.observe(observer, 'value')
    show_inline_matplotlib_plots()
    observer(None)
    return out


class Timer:
    def __init__(self, timeout, callback):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())

    async def _job(self):
        await asyncio.sleep(self._timeout)
        self._callback()

    def cancel(self):
        self._task.cancel()


def debounce(wait):
    """ Decorator that will postpone a function's
        execution until after `wait` seconds
        have elapsed since the last time it was invoked. """
    def decorator(fn):
        timer = None
        def debounced(*args, **kwargs):
            nonlocal timer
            def call_it():
                fn(*args, **kwargs)
            if timer is not None:
                timer.cancel()
            timer = Timer(wait, call_it)
        return debounced
    return decorator


import numpy as np
import matplotlib.pyplot as plt
from ipywidgets import widgets
from scipy.signal import stft
from pynwb.ecephys import LFP, SpikeEventSeries
from .base import fig2widget, nwb2widget


def show_lfp(node: LFP, neurodata_vis_spec: dict):
    lfp = list(node.electrical_series.values())[0]
    return nwb2widget(lfp, neurodata_vis_spec)


def show_spectrogram(neurodata, channel=0, **kwargs):
    fig, ax = plt.subplots()
    f, t, Zxx = stft(neurodata.data[:, channel], neurodata.rate, nperseg=2*17)
    ax.imshow(np.log(np.abs(Zxx)), aspect='auto', extent=[0, max(t), 0, max(f)], origin='lower')
    ax.set_ylim(0, max(f))
    ax.set_xlabel('time')
    ax.set_ylabel('frequency')
    plt.show(ax)


def show_spike_event_series(ses: SpikeEventSeries, **kwargs):
    def control_plot(spk_ind):
        fig, ax = plt.subplots(figsize=(9, 5))
        data = ses.data[spk_ind]
        if nChannels > 1:
            for ch in range(nChannels):
                ax.plot(data[:, ch], color='#d9d9d9')
        else:
            ax.plot(data[:], color='#d9d9d9')
        ax.plot(np.mean(data, axis=1), color='k')
        ax.set_xlabel('Time')
        ax.set_ylabel('Amplitude')
        plt.show()
        return fig2widget(fig)

    if len(ses.data.shape) == 3:
        nChannels = ses.data.shape[2]
    else:
        nChannels = ses.data.shape[1]
    nSpikes = ses.data.shape[0]

    # Controls
    field_lay = widgets.Layout(max_height='40px', max_width='100px',
                               min_height='30px', min_width='70px')
    spk_ind = widgets.BoundedIntText(value=0, min=0, max=nSpikes-1,
                                     layout=field_lay)
    controls = {'spk_ind': spk_ind}
    out_fig = widgets.interactive_output(control_plot, controls)

    # Assemble layout box
    lbl_spk = widgets.Label('Spike ID:', layout=field_lay)
    lbl_nspks0 = widgets.Label('N° spikes:', layout=field_lay)
    lbl_nspks1 = widgets.Label(str(nSpikes), layout=field_lay)
    lbl_nch0 = widgets.Label('N° channels:', layout=field_lay)
    lbl_nch1 = widgets.Label(str(nChannels), layout=field_lay)
    hbox0 = widgets.HBox(children=[lbl_spk, spk_ind])
    vbox0 = widgets.VBox(children=[
        widgets.HBox(children=[lbl_nspks0, lbl_nspks1]),
        widgets.HBox(children=[lbl_nch0, lbl_nch1]),
        hbox0
    ])
    hbox1 = widgets.HBox(children=[vbox0, out_fig])

    return hbox1



import numpy as np
from pynwb import TimeSeries
from datetime import datetime
from dateutil.tz import tzlocal
from pynwb import NWBFile
from pynwb.epoch import TimeIntervals
from nwbwidgets.utils.timeseries import get_timeseries_tt,get_timeseries_maxt,get_timeseries_mint,get_timeseries_in_units, \
timeseries_time_to_ind,align_by_times,align_by_trials,align_by_trials,align_by_time_intervals
import unittest


def test_get_timeseries_tt():
    
    data = list(range(100, 200, 10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0.0, rate=1.0)
    
    tt = get_timeseries_tt(ts)
    np.testing.assert_array_equal(tt,[0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    
def test_get_timeseries_tt_infstarting_time():
    
    data = list(range(100, 200, 10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=np.inf, rate=1.0)
    
    tt = get_timeseries_tt(ts)
    np.testing.assert_array_equal(tt,[0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    
def test_get_timeseries_tt_negativeistop():
    
    data = list(range(100, 200, 10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0., rate=1.0)
    
    tt = get_timeseries_tt(ts,istop=-1)
    np.testing.assert_array_equal(tt,[0., 1., 2., 3., 4., 5., 6., 7.])

    
def test_get_timeseries_in_units():
    data = list(range(100, 200, 10))
    timestamps = list(range(10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', timestamps=timestamps,conversion=np.inf)
    data,unit = get_timeseries_in_units(ts)
    assert(unit is None)
    assert(data==[100, 110, 120, 130, 140, 150, 160, 170, 180, 190])
    
    
def test_align_by_trials():
    start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
    create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

    nwbfile = NWBFile(session_description='NWBFile for PSTH', 
                      identifier='NWB123',  
                      session_start_time=start_time,  
                      file_create_date=create_date)
    
    data = list(range(100, 200, 10))
    timestamps = list(range(10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', timestamps=timestamps)
    nwbfile.add_acquisition(ts)
    
    nwbfile.add_trial_column(name='stim', description='the visual stimuli during the trial')
    nwbfile.add_trial(start_time=0.0, stop_time=2.0, stim='person')
    nwbfile.add_trial(start_time=3.0, stop_time=5.0, stim='ocean')
    nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim='desert')
    
    np.testing.assert_array_equal(align_by_trials(ts),np.array([[110],[140],[170]]))
    
    
    
     
class TimeSeriesTimeStampTestCase(unittest.TestCase):

    def setUp(self):
        data = list(range(100, 200, 10))
        timestamps = list(range(10))
        self.ts = TimeSeries(name='test_timeseries', data=data, unit='m', timestamps=timestamps)
    
    def test_get_timeseries_maxt(self):
        maxt = get_timeseries_maxt(self.ts)
        assert(maxt==9)
      
    def test_get_timeseries_mint(self):
        mint = get_timeseries_mint(self.ts)
        assert(mint==0)

    def test_timeseries_time_to_ind(self):
        assert(timeseries_time_to_ind(self.ts,3)==4)

    def test_align_by_times(self):
        assert(np.array_equal(align_by_times(self.ts,[0,1,2],[4,5,6]),np.array([[110, 120, 130, 140],[120, 130, 140, 150],[130, 140, 150, 160]])))
    
    def test_get_timeseries_tt_timestamp(self):
        tt = get_timeseries_tt(self.ts)
        np.testing.assert_array_equal(tt,[0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
        
    def test_align_by_time_intervals(self):
        intervals=TimeIntervals(name='Time Intervals')
        np.testing.assert_array_equal(align_by_time_intervals(timeseries=self.ts,intervals=intervals,stop_label=None),np.array([]))
        
    



from nwbwidgets import view
from ipywidgets import widgets
from collections.abc import Iterable
from datetime import datetime
from pynwb.file import NWBFile
from .base import lazy_show_over_data
from functools import partial


def show_nwbfile(nwbfile: NWBFile, neurodata_vis_spec: dict) -> widgets.Widget:
    """
        Gets a pynwb object and returns a Vertical Box containing textual info and
        an expandable Accordion with it's children.
    """
    field_lay = widgets.Layout(max_height='40px', max_width='500px',
                               min_height='30px', min_width='180px')
    info = []  # string data type, exposed as a Text widget
    neuro_data = []  # more complex data types, also with children
    labels = []
    for key, value in nwbfile.fields.items():
        if isinstance(value, (str, datetime)):
            lbl_key = widgets.Label(key + ':', layout=field_lay)
            lbl_val = widgets.Label(str(value), layout=field_lay)
            info.append(widgets.HBox(children=[lbl_key, lbl_val]))
        elif key == 'related_publications':
            pub_list = []
            for pub in value:
                if isinstance(pub, bytes):
                    pub = pub.decode()
                pub_list.append(widgets.HTML(value="<a href=http://dx.doi.org/" + pub[4:] + ">" + pub + "</a>"))
            lbl_key = widgets.Label(key + ':', layout=field_lay)
            pub_list.insert(0, lbl_key)
            info.append(widgets.HBox(children=pub_list))
        elif key == 'experimenter':
            lbl_experimenter = widgets.Label('Experimenter:', layout=field_lay)
            if isinstance(value, (list, tuple)):
                if isinstance(value[0], str):
                    lbl_names = widgets.Label(', '.join(value), layout=field_lay)
                elif isinstance(value[0], bytes):
                    lbl_names = widgets.Label(b', '.join(value).decode(), layout=field_lay)
                else:
                    raise ValueError('unrecognized type for experimenter')
            else:
                lbl_names = widgets.Label(value, layout=field_lay)
            hbox_exp = widgets.HBox(children=[lbl_experimenter, lbl_names])
            info.append(hbox_exp)
        elif (isinstance(value, Iterable) and len(value)) or value:
            neuro_data.append(value)
            if hasattr(nwbfile.fields[key], 'description') and nwbfile.fields[key].description:
                labels.append(key + ': ' + nwbfile.fields[key].description)
            else:
                labels.append(key)
    func_ = partial(view.nwb2widget, neurodata_vis_spec=neurodata_vis_spec)
    accordion = lazy_show_over_data(neuro_data, func_, labels=labels, style=widgets.Accordion)

    return widgets.VBox(info + [accordion])


import numpy as np
import pynwb

from bisect import bisect_right, bisect_left
from numpy import searchsorted


def get_spike_times(units: pynwb.misc.Units, index, in_interval):
    """Use bisect methods to efficiently retrieve spikes from a given unit in a given interval

    Parameters
    ----------
    units: pynwb.misc.Units
    index: int
    in_interval: start and stop times

    Returns
    -------

    """
    st = units['spike_times']
    unit_start = 0 if index == 0 else st.data[index - 1]
    unit_stop = st.data[index]
    start_time, stop_time = in_interval

    ind_start = bisect_left(st.target, start_time, unit_start, unit_stop)
    ind_stop = bisect_right(st.target, stop_time, ind_start, unit_stop)

    return np.asarray(st.target[ind_start:ind_stop])


def get_min_spike_time(units: pynwb.misc.Units):
    """Efficiently retrieve the first spike time across all units

    Parameters
    ----------
    units: pynwb.misc.Units

    Returns
    -------

    """
    st = units['spike_times']
    inds = [0] + list(st.data[:-1])
    first_spikes = [st.target.data[i] for i in inds]
    return np.min(first_spikes)


def get_max_spike_time(units: pynwb.misc.Units):
    """Efficiently retrieve the last spike time across all units

    Parameters
    ----------
    units: pynwb.misc.Units

    Returns
    -------

    """
    st = units['spike_times']
    inds = [x - 1 for x in st.data[:]]
    last_spikes = [st.target.data[i] for i in inds]
    return np.max(last_spikes)


def align_by_times(units: pynwb.misc.Units, index, starts, stops):
    """
    Args:
        units: pynwb.misc.Units
        index: int
        starts: array-like
        stops: array-like
    Returns:
        np.array
    """

    st = units['spike_times']
    unit_spike_data = st[index]

    istarts = searchsorted(unit_spike_data, starts)
    istops = searchsorted(unit_spike_data, stops)
    for start, istart, istop in zip(starts, istarts, istops):
        yield unit_spike_data[istart:istop] - start


def align_by_trials(units: pynwb.misc.Units, index, start_label='start_time',
                    stop_label=None, before=0., after=1.):
    """
    Args:
        units
        start_label: str
            default: 'start_time'
        stop_label: str
            default: None (just align to start_time)
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    trials = units.get_ancestor('NWBFile').trials
    return align_by_time_intervals(units, index, trials, start_label, stop_label, before, after)


def align_by_time_intervals(units: pynwb.misc.Units, index, intervals, start_label='start_time',
                            stop_label='stop_time', before=0., after=0., rows_select=(), progress_bar=None):
    """
    Args:
        units: time-aware neurodata_type
        index: int
        intervals: pynwb.epoch.TimeIntervals
        start_label: str
            default: 'start_time'
        stop_label: str
            default: 'stop_time'
        before: float
            time after start_label in secs (positive goes back in time)
        after: float
            time after stop_label in secs (positive goes forward in time)
        rows_select: array_like, optional
            sub-selects specific rows
        progress_bar: FloatProgress, optional
            Proved a progress bar object to have this method automatically update the progress bar
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    if stop_label is None:
        stop_label = start_label
    starts = np.array(intervals[start_label][:])[rows_select] - before
    stops = np.array(intervals[stop_label][:])[rows_select] + after
    if progress_bar is not None:
        progress_bar.value = 0
        progress_bar.description = 'reading spike data'

    out = []
    for i, x in enumerate(align_by_times(units, index, starts, stops)):
        out.append(x - before)
        if progress_bar is not None:
            progress_bar.value = i / len(units)

    return out


def get_unobserved_intervals(units, time_window, units_select=()):

    if 'obs_intervals' not in units:
        return []

    # add observation intervals
    unobserved_intervals_list = []
    for i_unit in units_select:
        intervals = units['obs_intervals'][i_unit]  # TODO: use bisect here
        intervals = np.array(intervals)
        these_obs_intervals = intervals[(intervals[:, 1] > time_window[0]) & (intervals[:, 0] < time_window[1])]
        unobs_intervals = np.c_[these_obs_intervals[:-1, 1], these_obs_intervals[1:, 0]]

        if len(these_obs_intervals):
            # handle unobserved interval on lower bound of window
            if these_obs_intervals[0, 0] > time_window[0]:
                unobs_intervals = np.vstack(([time_window[0], these_obs_intervals[0, 0]], unobs_intervals))

            # handle unobserved interval on lower bound of window
            if these_obs_intervals[-1, 1] < time_window[1]:
                unobs_intervals = np.vstack((unobs_intervals, [these_obs_intervals[-1, 1], time_window[1]]))
        else:
            unobs_intervals = [time_window]

        unobserved_intervals_list.append(unobs_intervals)

    return unobserved_intervals_list



import numpy as np
from pynwb import TimeSeries
from pynwb.behavior import Position, SpatialSeries, BehavioralEvents
from nwbwidgets.view import default_neurodata_vis_spec
from nwbwidgets.behavior import show_position, show_behavioral_events, show_spatial_series_over_time, show_spatial_series
import unittest

class ShowSpatialSeriesTestCase(unittest.TestCase):

    def setUp(self):
        self.spatial_series = SpatialSeries(name = 'position',
                                   data = np.linspace(0, 1, 20),
                                   rate = 50.,
                                   reference_frame = 'starting gate')
    def test_show_position(self):
 
        position = Position(spatial_series = self.spatial_series)

        show_position(position, default_neurodata_vis_spec)

    def test_show_spatial_series_over_time(self):

        show_spatial_series_over_time(self.spatial_series)

    def test_show_spatial_series(self):

        show_spatial_series(self.spatial_series)

        
class ShowSpatialSeriesTwoDTestCase(unittest.TestCase):

    def setUp(self):
        self.spatial_series = SpatialSeries(name = 'position',
                                       data = np.array([np.linspace(0, 1, 20), np.linspace(0, 1, 20)]).T,
                                       rate = 50.,
                                       reference_frame = 'starting gate')

    def test_show_spatial_series_over_time_twoD(self):
        
        show_spatial_series_over_time(self.spatial_series)

    def test_show_spatial_series_twoD(self):

        show_spatial_series(self.spatial_series)
    

class ShowSpatialSeriesThreeDTestCase(unittest.TestCase):

    def setUp(self):
        self.spatial_series = SpatialSeries(name = 'position',
                                       data = np.array([np.linspace(0, 1, 20), np.linspace(0, 1, 20),np.linspace(0, 1, 20)]).T,
                                       rate = 50.,
                                       reference_frame = 'starting gate')

    def test_show_spatial_series_over_time_threeD(self):
        
        show_spatial_series_over_time(self.spatial_series)

    def test_show_spatial_series_threeD(self):

        show_spatial_series(self.spatial_series)

        
def test_show_behavioral_events():
    
    data = list(range(100, 200, 10))
    ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0.0, rate=1.0)
    
    beh_events = BehavioralEvents(time_series=ts)
    
    show_behavioral_events(beh_events, default_neurodata_vis_spec)



from setuptools import setup


with open('README.md') as f:
    long_description = f.read()

setup(
    author='Neurodata Without Border (NWB) developers',
    author_email='ben.dichter@gmail.com',
    version='0.2.1',
    classifiers=['Operating System :: OS Independent',
                 'Development Status :: 3 - Alpha',
                 'Framework :: Jupyter',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 ],
    description='This is nwbwidgets, widgets for viewing the contents of a '
                'NWB-file in Jupyter Notebooks using ipywidgets.',
    install_requires=['pynwb',
                      'ipympl',
                      'ipywidgets>=7.4',
                      'matplotlib',
                      'numpy',
                      'ipyvolume',
                      'ndx_grayscalevolume',
                      'plotly',
                      'scikit-image',
                      'tqdm',
                      'ndx-icephys-meta'],
    license='MIT',
    keywords=['jupyter', 'hdf5', 'notebook', 'nwb'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='nwbwidgets',
    packages=['nwbwidgets', 'nwbwidgets/utils'],
    python_requires='>=2.7',
    setup_requires=['setuptools>=38.6.0', 'setuptools_scm'],
    url='https://github.com/NeurodataWithoutBorders/nwb-jupyter-widgets')



"""
Taken from https://github.com/maartenbreddels/ipyvolume/pull/178
"""
import numpy as np
import matplotlib.cm
import matplotlib.colors
from ipyvolume import TransferFunction


def linear_transfer_function(color,
                             min_opacity=0,
                             max_opacity=0.05,
                             reverse_opacity=False,
                             n_elements=256):
    """Transfer function of a single color and linear opacity.
    :param color: List-like RGB, or string with hexadecimal or named color.
        RGB values should be within 0-1 range.
    :param min_opacity: Minimum opacity, default value is 0.0.
        Lowest possible value is 0.0, optional.
    :param max_opacity: Maximum opacity, default value is 0.05.
        Highest possible value is 1.0, optional.
    :param reverse_opacity: Linearly decrease opacity, optional.
    :param n_elements: Length of rgba array transfer function attribute.
    :type color: list-like or string
    :type min_opacity: float, int
    :type max_opacity: float, int
    :type reverse_opacity: bool
    :type n_elements: int
    :return: transfer_function
    :rtype: ipyvolume TransferFunction
    :Example:
    >>> import ipyvolume as ipv
    >>> green_tf = ipv.transfer_function.linear_transfer_function('green')
    >>> ds = ipv.datasets.aquariusA2.fetch()
    >>> ipv.volshow(ds.data[::4,::4,::4], tf=green_tf)
    >>> ipv.show()
    .. seealso:: matplotlib_transfer_function()
    """
    r, g, b = matplotlib.colors.to_rgb(color)
    opacity = np.linspace(min_opacity, max_opacity, num=n_elements)
    if reverse_opacity:
        opacity = np.flip(opacity, axis=0)
    rgba = np.transpose(np.stack([[r] * n_elements,
                                  [g] * n_elements,
                                  [b] * n_elements,
                                  opacity]))
    transfer_function = TransferFunction(rgba=rgba)
    return transfer_function


def matplotlib_transfer_function(colormap_name,
                                 min_opacity=0,
                                 max_opacity=0.05,
                                 reverse_colormap=False,
                                 reverse_opacity=False,
                                 n_elements=256):
    """Transfer function from matplotlib colormaps.
    :param colormap_name: name of matplotlib colormap
    :param min_opacity: Minimum opacity, default value is 0.
        Lowest possible value is 0, optional.
    :param max_opacity: Maximum opacity, default value is 0.05.
        Highest possible value is 1.0, optional.
    :param reverse_colormap: reversed matplotlib colormap, optional.
    :param reverse_opacity: Linearly decrease opacity, optional.
    :param n_elements: Length of rgba array transfer function attribute.
    :type colormap_name: str
    :type min_opacity: float, int
    :type max_opacity: float, int
    :type reverse_colormap: bool
    :type reverse_opacity: bool
    :type n_elements: int
    :return: transfer_function
    :rtype: ipyvolume TransferFunction
    :Example:
    >>> import ipyvolume as ipv
    >>> rgb = (0, 255, 0)  # RGB value for green
    >>> green_tf = ipv.transfer_function.matplotlib_transfer_function('bone')
    >>> ds = ipv.datasets.aquariusA2.fetch()
    >>> ipv.volshow(ds.data[::4,::4,::4], tf=green_tf)
    >>> ipv.show()
    .. seealso:: linear_transfer_function()
    """
    cmap = matplotlib.cm.get_cmap(name=colormap_name)
    rgba = np.array([cmap(i) for i in np.linspace(0, 1, n_elements)])
    if reverse_colormap:
        rgba = np.flip(rgba, axis=0)
    # Create opacity values to overwrite default matplotlib opacity=1.0
    opacity = np.linspace(min_opacity, max_opacity, num=n_elements)
    if reverse_opacity:
        opacity = np.flip(opacity, axis=0)
    rgba[:, -1] = opacity  # replace opacity=1 with actual opacity
    transfer_function = TransferFunction(rgba=rgba)
    return transfer_function



from pynwb.core import DynamicTable
import numpy as np


def infer_categorical_columns(dynamic_table: DynamicTable):

    categorical_cols = {}
    for name in dynamic_table.colnames:
        if len(dynamic_table[name].shape) == 1:
            try:  # TODO: fix this
                unique_vals = np.unique(dynamic_table[name].data)
                if 1 < len(unique_vals) <= (len(dynamic_table[name].data) / 2):
                    categorical_cols[name] = unique_vals
            except:
                pass
    return categorical_cols


def group_and_sort(group_vals=None, group_select=None, order_vals=None, discard_rows=None, limit=None):
    """
    Logical flow:
    0) Apply discard_rows - throw out any listed rows
    1) Apply group select - Return only values that are within this group
    2) Apply order - If group is provided, items are sorted first by group and then by order
    3) Apply limit - Applied per group
    4) Apply window - Return only items [int] through [int]. Useful for plotting

    Parameters
    ----------
    group_vals: array-like of str
        ['a', 'b', 'b', 'a', 'a', 'b', 'b', 'c']
    group_select: array-like of str
        ['a', 'b']
    order_vals: array-like of ints
        [0, 3, 4, 1, 2, 5, 6, 7]
    limit: int

    Returns
    -------

    """

    if group_vals is not None:
        if order_vals is None:
            order = np.argsort(group_vals)
        else:
            order = np.lexsort([order_vals, group_vals])
        labels, group_inds = np.unique(group_vals[order], return_inverse=True)
    else:
        labels, group_inds = None, None
        if order_vals is None:
            # cannot give order because we don't know how many units there are
            raise ValueError('group_vals and order_vals cannot both be None')
        else:
            order = np.argsort(order_vals)

    # apply discard rows
    if discard_rows is not None:
        keep = np.logical_not(np.isin(order, discard_rows))
        order = order[keep]
        if group_inds is not None:
            group_inds = group_inds[keep]

    # apply discard NaN categories
    try:
        if any(np.isnan(labels)):
            nan_labs = np.isnan(labels)
            keep = ~np.isin(group_inds, np.where(nan_labs)[0])
            group_inds = group_inds[keep]
            order = order[keep]
            labels = labels[~np.isnan(labels)]
    except TypeError:  # if labels are strings
        pass

    # apply discard groups (but keep labels)
    if group_select is not None:
        keep = np.isin(labels[group_inds], group_select)
        group_inds = group_inds[keep]
        order = order[keep]

    # remove groups that are missing
    if labels is not None:
        labels = labels[np.isin(range(len(labels)), group_inds)]
        _, group_inds = np.unique(group_inds, return_inverse=True)

    # apply limit
    inds = list()
    if limit is not None:
        if group_inds is not None:
            for i in range(len(labels)):
                inds.extend(np.where(group_inds == i)[0][:limit])
            order = order[inds]
            group_inds = group_inds[inds]
        else:
            order = order[:limit]

    return order, group_inds, labels



import numpy as np
import matplotlib.pyplot as plt
from pynwb import TimeSeries
from ipywidgets import widgets
from nwbwidgets.timeseries import traces_widget, show_ts_fields, show_timeseries, plot_traces, show_timeseries_mpl
import unittest


def test_timeseries_widget():
    ts = TimeSeries(name='name', description='no description',
                    data=np.array([[1., 2., 3., 4.],
                                   [11., 12., 13., 14.]]),
                    rate=100.)

    traces_widget(ts)


class ShowTimeSeriesTestCase(unittest.TestCase):
    
    def setUp(self):
        data = np.random.rand(160,3)
        self.ts = TimeSeries(name='test_timeseries', data=data, unit='m', starting_time=0.0, rate=1.0)

    def test_show_ts_fields(self):
        assert isinstance(show_ts_fields(self.ts), widgets.Widget)
        
    def test_show_timeseries(self):
        assert isinstance(show_timeseries(self.ts, istart=5, istop=56), widgets.Widget)
    
    def test_show_timeseries_mpl(self):
        ax = show_timeseries_mpl(self.ts, zero_start=True, title='Test show_timeseries_mpl')
        assert isinstance(ax, plt.Subplot)


class PlotTracesTestCase(unittest.TestCase):
    
    def setUp(self):
        self.data = np.random.rand(160, 3)
    
    def test_plot_traces(self):

        ts = TimeSeries(name='test_timeseries', data=self.data, unit='m', starting_time=0.0, rate=20.0)
        plot_traces(ts)
        
    def test_plot_traces_fix(self):

        ts = TimeSeries(name='test_timeseries', data=self.data.T, unit='m', starting_time=0.0, rate=20.0)
        plot_traces(ts)
